import napari
from skimage import data, morphology
from skan import Skeleton
from skan.napari_skan import labels_to_skeleton_shapes, SkeletonizeMethod
import numpy as np

viewer = napari.Viewer()
horse = np.logical_not(data.horse().astype(bool))

labels_layer = viewer.add_labels(horse)

ldt = labels_to_skeleton_shapes(labels_layer, SkeletonizeMethod.zhang)
(skel_layer,) = viewer._add_layer_from_data(*ldt)

dw, widget = viewer.window.add_plugin_dock_widget(
        'skan', 'Color Skeleton Widget'
        )

napari.run()



import numpy as np
import pandas as pd

from skan import io

from skimage._shared._tempfile import temporary_file


def test_write_excel_tables():
    num_sheets = np.random.randint(1, 4)
    num_cols = np.random.randint(1, 5, size=num_sheets)
    num_rows = np.random.randint(20, 40, size=num_sheets)
    tables = []
    for m, n in zip(num_rows, num_cols):
        columns = [f'column{i}' for i in range(n)]
        data = np.random.random((m, n))
        tables.append(pd.DataFrame(data=data, columns=columns))
    sheet_names = [f'sheet {i}' for i in range(num_sheets)]
    kwargs = dict(zip(sheet_names, tables))
    kwargs['config'] = {
            'image files': ['image1.tif', 'image2.tif'], 'image format': 'fei',
            'threshold radius': 5e-8
            }
    with temporary_file(suffix='.xlsx') as file:
        io.write_excel(file, **kwargs)
        tables_in = [
                pd.read_excel(
                        file, sheet_name=name, index_col=0, engine='openpyxl'
                        ) for name in sheet_names
                ]
        config_in_df = pd.read_excel(
                file, sheet_name='config', engine='openpyxl'
                )
        config_in = dict(
                zip(config_in_df['parameters'], config_in_df['values'])
                )
    for table, table_in in zip(tables, tables_in):
        assert list(table.columns) == list(table_in.columns)
        np.testing.assert_allclose(table_in.values, table.values)
    for key, val in kwargs['config'].items():
        str(val) == str(config_in[key])



from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property

import networkx as nx
import numpy as np
import pandas as pd
import scipy
from scipy import sparse, ndimage as ndi
from scipy.sparse import csgraph
from scipy.spatial import distance_matrix
from skimage import morphology
from skimage.graph import central_pixel
from skimage.util._map_array import map_array, ArrayMap
import numpy.typing as npt
import numba
import warnings
from typing import Tuple, Callable
from .nputil import _raveled_offsets_and_distances
from .summary_utils import find_main_branches


def _weighted_abs_diff(
        values0: np.ndarray, values1: np.ndarray, distances: np.ndarray
        ) -> np.ndarray:
    """A default edge function for complete image graphs.

    A pixel graph on an image with no edge values and no mask is a very
    boring regular lattice, so we define a default edge weight to be the
    absolute difference between values *weighted* by the distance
    between them.

    Parameters
    ----------
    values0 : array
        The pixel values for each node.
    values1 : array
        The pixel values for each neighbor.
    distances : array
        The distance between each node and its neighbor.

    Returns
    -------
    edge_values : array of float
        The computed values: abs(values0 - values1) * distances.
    """
    return np.abs(values0 - values1) * distances


def pixel_graph(
        image, *, mask=None, edge_function=None, connectivity=1, spacing=None
        ):
    """Create an adjacency graph of pixels in an image.

    Pixels where the mask is True are nodes in the returned graph, and they are
    connected by edges to their neighbors according to the connectivity
    parameter. By default, the *value* of an edge when a mask is given, or when
    the image is itself the mask, is the euclidean distance betwene the pixels.

    However, if an int- or float-valued image is given with no mask, the value
    of the edges is the absolute difference in intensity between adjacent
    pixels, weighted by the euclidean distance.

    Parameters
    ----------
    image : array
        The input image. If the image is of type bool, it will be used as the
        mask as well.
    mask : array of bool
        Which pixels to use. If None, the graph for the whole image is used.
    edge_function : callable
        A function taking an array of pixel values, and an array of neighbor
        pixel values, and an array of distances, and returning a value for the
        edge. If no function is given, the value of an edge is just the
        distance.
    connectivity : int
        The square connectivity of the pixel neighborhood: the number of
        orthogonal steps allowed to consider a pixel a neigbor. See
        `scipy.ndimage.generate_binary_structure` for details.
    spacing : tuple of float
        The spacing between pixels along each axis.

    Returns
    -------
    graph : scipy.sparse.csr_matrix
        A sparse adjacency matrix in which entry (i, j) is 1 if nodes i and j
        are neighbors, 0 otherwise.
    nodes : array of int
        The nodes of the graph. These correspond to the raveled indices of the
        nonzero pixels in the mask.
    """
    if image.dtype == bool and mask is None:
        mask = image
    if mask is None and edge_function is None:
        mask = np.ones_like(image, dtype=bool)
        edge_function = _weighted_abs_diff

    # Strategy: we are going to build the (i, j, data) arrays of a scipy
    # sparse COO matrix, then convert to CSR (which is fast).
    # - grab the raveled IDs of the foreground (mask == True) parts of the
    #   image **in the padded space**.
    # - broadcast them together with the raveled offsets to their neighbors.
    #   This gives us for each foreground pixel a list of neighbors (that
    #   may or may not be selected by the mask.) (We also track the *distance*
    #   to each neighbor.)
    # - select "valid" entries in the neighbors and distance arrays by indexing
    #   into the mask, which we can do since these are raveled indices.
    # - use np.repeat() to repeat each source index according to the number
    #   of neighbors selected by the mask it has. Each of these repeated
    #   indices will be lined up with its neighbor, i.e. **this is the i
    #   array** of the COO format matrix.
    # - use the mask as a boolean index to get a 1D view of the selected
    #   neighbors. **This is the j array.**
    # - by default, the same boolean indexing can be applied to the distances
    #   to each neighbor, to give the **data array.** Optionally, a
    #   provided edge function can be computed on the pixel values and the
    #   distances to give a different value for the edges.
    # Note, we use map_array to map the raveled coordinates in the padded
    # image to the ones in the original image, and those are the returned
    # nodes.
    padded = np.pad(mask, 1, mode='constant', constant_values=False)
    nodes_padded = np.flatnonzero(padded)
    neighbor_offsets_padded, distances_padded = _raveled_offsets_and_distances(
            padded.shape, connectivity=connectivity, spacing=spacing
            )
    neighbors_padded = nodes_padded[:, np.newaxis] + neighbor_offsets_padded
    neighbor_distances_full = np.broadcast_to(
            distances_padded, neighbors_padded.shape
            )
    nodes = np.flatnonzero(mask)
    nodes_sequential = np.arange(nodes.size)
    # neighbors outside the mask get mapped to 0, which is a valid index,
    # BUT, they will be masked out in the next step.
    neighbors = map_array(neighbors_padded, nodes_padded, nodes)
    neighbors_mask = padded.reshape(-1)[neighbors_padded]
    num_neighbors = np.sum(neighbors_mask, axis=1)
    indices = np.repeat(nodes, num_neighbors)
    indices_sequential = np.repeat(nodes_sequential, num_neighbors)
    neighbor_indices = neighbors[neighbors_mask]
    neighbor_distances = neighbor_distances_full[neighbors_mask]
    neighbor_indices_sequential = map_array(
            neighbor_indices, nodes, nodes_sequential
            )
    if edge_function is None:
        data = neighbor_distances
    else:
        image_r = image.reshape(-1)
        data = edge_function(
                image_r[indices], image_r[neighbor_indices], neighbor_distances
                )
    m = nodes_sequential.size
    mat = sparse.coo_matrix(
            (data, (indices_sequential, neighbor_indices_sequential)),
            shape=(m, m)
            )
    graph = mat.tocsr()
    return graph, nodes


csr_spec_float = [
        ('indptr', numba.int32[:]),
        ('indices', numba.int32[:]),
        ('data', numba.float64[:]),
        ('shape', numba.int32[:]),
        ('node_properties', numba.float64[:]),
        ]  # yapf: disable

csr_spec_bool = [
        ('indptr', numba.int32[:]),
        ('indices', numba.int32[:]),
        ('data', numba.bool_[:]),
        ('shape', numba.int32[:]),
        ('node_properties', numba.float64[:]),
        ]  # yapf: disable


class NBGraphBase:

    def __init__(self, indptr, indices, data, shape, node_props):
        self.indptr = indptr
        self.indices = indices
        self.data = data
        self.shape = shape
        self.node_properties = node_props

    def edge(self, i, j):
        return _csrget(self.indices, self.indptr, self.data, i, j)

    def set_edge(self, i, j, value):
        return _csrset(self.indices, self.indptr, self.data, i, j, value)

    def neighbors(self, row):
        loc, stop = self.indptr[row], self.indptr[row + 1]
        return self.indices[loc:stop]

    @property
    def has_node_props(self):
        return self.node_properties.strides != (0,)


NBGraph = numba.experimental.jitclass(NBGraphBase, csr_spec_float)
NBGraphBool = numba.experimental.jitclass(NBGraphBase, csr_spec_bool)


def csr_to_nbgraph(csr, node_props=None):
    if node_props is None:
        node_props = np.broadcast_to(1., csr.shape[0])
        node_props.flags.writeable = True
    return NBGraph(
            csr.indptr.astype(np.int32, copy=False),
            csr.indices.astype(np.int32, copy=False),
            csr.data,
            np.array(csr.shape, dtype=np.int32),
            node_props.astype(np.float64),
            )


def _pixel_graph(image, steps, distances, num_edges, height=None):
    row = np.empty(num_edges, dtype=int)
    col = np.empty(num_edges, dtype=int)
    data = np.empty(num_edges, dtype=float)
    if height is None:
        k = _write_pixel_graph(image, steps, distances, row, col, data)
    else:
        k = _write_pixel_graph_height(
                image, height, steps, distances, row, col, data
                )
    graph = sparse.coo_matrix((data[:k], (row[:k], col[:k]))).tocsr()
    return graph


@numba.jit(nopython=True, cache=True, nogil=True)
def _write_pixel_graph(image, steps, distances, row, col, data):
    """Step over `image` to build a graph of nonzero pixel neighbors.

    Parameters
    ----------
    image : int array
        The input image.
    steps : int array, shape (N,)
        The raveled index steps to find a pixel's neighbors in `image`.
    distances : float array, shape (N,)
        The euclidean distance from a pixel to its corresponding
        neighbor in `steps`.
    row : int array
        Output array to be filled with the "center" pixel IDs.
    col : int array
        Output array to be filled with the "neighbor" pixel IDs.
    data : float array
        Output array to be filled with the distances from center to
        neighbor pixels.

    Returns
    -------
    k : int
        The number of entries written to row, col, and data.

    Notes
    -----
    No size or bounds checking is performed. Users should ensure that
    - No index in `indices` falls on any edge of `image` (or the
      neighbor computation will fail or segfault).
    - The `steps` and `distances` arrays have the same shape.
    - The `row`, `col`, `data` are long enough to hold all of the
      edges.
    """
    image = image.ravel()
    n_neighbors = steps.size
    start_idx = np.max(steps)
    end_idx = image.size + np.min(steps)
    k = 0
    for i in range(start_idx, end_idx + 1):
        if image[i] != 0:
            for j in range(n_neighbors):
                n = steps[j] + i
                if image[n] != 0 and image[n] != image[i]:
                    row[k] = image[i]
                    col[k] = image[n]
                    data[k] = distances[j]
                    k += 1
    return k


@numba.jit(nopython=True, cache=True, nogil=True)
def _write_pixel_graph_height(image, height, steps, distances, row, col, data):
    """Step over `image` to build a graph of nonzero pixel neighbors.

    Parameters
    ----------
    image : int array
        The input image.
    height : float array, same shape as `image`
        This is taken to be a height map along an additional
        dimension (in addition to the image dimensions), so the distance
        between two neighbors `i` and `n` separated by `j` is given by:

             `np.sqrt(distances[j]**2 + (height[i] - height[n])**2)`

    steps : int array, shape (N,)
        The raveled index steps to find a pixel's neighbors in `image`.
    distances : float array, shape (N,)
        The euclidean distance from a pixel to its corresponding
        neighbor in `steps`.
    row : int array
        Output array to be filled with the "center" pixel IDs.
    col : int array
        Output array to be filled with the "neighbor" pixel IDs.
    data : float array
        Output array to be filled with the distances from center to
        neighbor pixels.

    Returns
    -------
    k : int
        The number of entries written to row, col, and data.

    Notes
    -----
    No size or bounds checking is performed. Users should ensure that
    - No index in `indices` falls on any edge of `image` (or the
      neighbor computation will fail or segfault).
    - The `steps` and `distances` arrays have the same shape.
    - The `row`, `col`, `data` are long enough to hold all of the
      edges.
    """
    image = image.ravel()
    height = height.ravel()
    n_neighbors = steps.size
    start_idx = np.max(steps)
    end_idx = image.size + np.min(steps)
    k = 0
    for i in range(start_idx, end_idx + 1):
        if image[i] != 0:
            for j in range(n_neighbors):
                n = steps[j] + i
                if image[n] != 0 and image[n] != image[i]:
                    row[k] = image[i]
                    col[k] = image[n]
                    data[k] = np.sqrt(
                            distances[j]**2 + (height[i] - height[n])**2
                            )
                    k += 1
    return k


@numba.jit(nopython=True, cache=False)  # change this to True with Numba 1.0
def _build_paths(jgraph, indptr, indices, path_data, visited, degrees):
    indptr_i = 0
    indices_j = 0
    # first, process all nodes in a path to an endpoint or junction
    for node in range(jgraph.shape[0]):
        if degrees[node] > 2 or degrees[node] == 1:
            for neighbor in jgraph.neighbors(node):
                if not visited.edge(node, neighbor):
                    n_steps = _walk_path(
                            jgraph,
                            node,
                            neighbor,
                            visited,
                            degrees,
                            indices,
                            path_data,
                            indices_j,
                            )
                    indptr[indptr_i + 1] = indptr[indptr_i] + n_steps
                    indptr_i += 1
                    indices_j += n_steps
    # everything else is by definition in isolated cycles
    for node in range(jgraph.shape[0]):
        if degrees[node] > 0:
            neighbor = jgraph.neighbors(node)[0]
            if not visited.edge(node, neighbor):
                n_steps = _walk_path(
                        jgraph,
                        node,
                        neighbor,
                        visited,
                        degrees,
                        indices,
                        path_data,
                        indices_j,
                        )
                indptr[indptr_i + 1] = indptr[indptr_i] + n_steps
                indptr_i += 1
                indices_j += n_steps
    return indptr_i + 1, indices_j


@numba.jit(nopython=True, cache=False)  # change this to True with Numba 1.0
def _walk_path(
        jgraph, node, neighbor, visited, degrees, indices, path_data, startj
        ):
    indices[startj] = node
    start_node = node
    path_data[startj] = jgraph.node_properties[node]
    j = startj + 1
    while not visited.edge(node, neighbor):
        visited.set_edge(node, neighbor, True)
        visited.set_edge(neighbor, node, True)
        indices[j] = neighbor
        path_data[j] = jgraph.node_properties[neighbor]
        if degrees[neighbor] != 2 or neighbor == start_node:
            break
        n1, n2 = jgraph.neighbors(neighbor)
        nextneighbor = n1 if n1 != node else n2
        node, neighbor = neighbor, nextneighbor
        j += 1
    return j - startj + 1


def _build_skeleton_path_graph(graph):
    max_num_cycles = graph.indices.size // 4
    buffer_size_offset = max_num_cycles
    degrees = np.diff(graph.indptr)
    visited_data = np.zeros(graph.data.shape, dtype=bool)
    visited = NBGraphBool(
            graph.indptr.astype(np.int32, copy=False),
            graph.indices.astype(np.int32, copy=False), visited_data,
            graph.shape, np.broadcast_to(1.0, graph.shape[0])
            )
    endpoints = (degrees != 2)
    endpoint_degrees = degrees[endpoints]
    num_paths = np.sum(endpoint_degrees)
    path_indptr = np.zeros(num_paths + buffer_size_offset, dtype=int)
    # the number of points that we need to save to store all skeleton
    # paths is equal to the number of pixels plus the sum of endpoint
    # degrees minus one (since the endpoints will have been counted once
    # already in the number of pixels) *plus* the number of isolated
    # cycles (since each cycle has one index repeated). We don't know
    # the number of cycles ahead of time, but it is bounded by one quarter
    # of the number of points.
    n_points = (
            graph.indices.size + np.sum(np.maximum(0, endpoint_degrees - 1))
            + buffer_size_offset
            )
    path_indices = np.zeros(n_points, dtype=int)
    path_data = np.zeros(path_indices.shape, dtype=float)
    m, n = _build_paths(
            graph, path_indptr, path_indices, path_data, visited, degrees
            )
    paths = sparse.csr_matrix(
            (path_data[:n], path_indices[:n], path_indptr[:m]),
            shape=(m - 1, n)
            )
    return paths


@dataclass
class PathGraph:
    """Generalization of a skeleton.

    A morphological skeleton is a collection of single pixel wide paths in
    a binary (or, more generally, quantitative) image. Skan was created to
    make measurements of those paths in image data. It turns out, though,
    that the neighborhood topology of those paths (a graph containing long
    strings of nodes of degree 2) is more generally useful, and indeed makes
    sense without a source image or without representing pixels. (One example:
    tracklets in cell tracking data.)

    In the text below, we use the following notation:

    - N: the number of points in the pixel skeleton,
    - ndim: the dimensionality of the skeleton
    - P: the number of paths in the skeleton (also the number of links in the
      junction graph).
    - J: the number of junction nodes
    - Sd: the sum of the degrees of all the junction nodes
    - [Nt], [Np], Nr, Nc: the dimensions of the source image
    """
    adj: scipy.sparse.csr_array  # pixel/node-id neighbor adjacency matrix
    node_coordinates: np.ndarray | None
    node_values: np.ndarray | None
    paths: scipy.sparse.csr_array  # paths[i, j] = 1 iff coord j is in path i
    spacing: float | tuple[float, ...] = 1  # spatial scale between coordinates

    @classmethod
    def from_graph(cls, *, adj, node_coordinates, node_values=None, spacing=1):
        """Build a PathGraph from an adjacency matrix and node coordinates.

        Parameters
        ----------
        adj : scipy.sparse.csr_array
            An adjacency matrix where adj[i, j] is nonzero iff there is an
            edge between node i and node j.
        node_coordinates : np.ndarray, shape (N, ndim)
            The coordinates of the nodes. node_coordinates[i] is the
            coordinate of node i. The indices of these coordinates must match
            the indices of adj.
        node_values : np.ndarray, shape (N,)
            Values of the nodes. Could be image intensity, height, or some
            other quantity of which you want to compute statistics along the
            path.
        spacing : float or tuple of float, shape (ndim,)
            The pixel/voxel spacing along each axis coordinate.
        """
        nbgraph = csr_to_nbgraph(adj, node_values)
        paths = _build_skeleton_path_graph(nbgraph)
        return cls(adj, node_coordinates, node_values, paths, spacing)

    @classmethod
    def from_image(cls, skeleton_image, *, spacing=1, value_is_height=False):
        """Build a PathGraph from a skeleton image.

        This is just a convenience meant to mirror Skeleton.__init__.
        """
        graph, coords = skeleton_to_csgraph(
                skeleton_image,
                spacing=spacing,
                value_is_height=value_is_height,
                )
        values = _extract_values(skeleton_image, coords)
        return cls.from_graph(
                adj=graph,
                node_coordinates=np.transpose(coords),
                node_values=values,
                spacing=spacing,
                )

    @cached_property
    def nbgraph(self):
        return csr_to_nbgraph(self.adj, self.node_values)

    @cached_property
    def distances(self):
        """The path distances.

        Returns
        -------
        distances : np.ndarray of float, shape (P,)
            distances[i] contains the distance of path i.
        """
        distances = np.empty(self.n_paths, dtype=float)
        _compute_distances(
                self.nbgraph, self.paths.indptr, self.paths.indices, distances
                )
        return distances

    @property
    def n_paths(self):
        return self.paths.shape[0]

    @cached_property
    def degrees(self):
        """The degree (number of neighbors) of each node/pixel.

        Returns
        -------
        degrees : np.ndarray of int, shape (N,)
        """
        return np.diff(self.adj.indptr)


def _extract_values(image, coords):
    if image.dtype == np.bool_:
        return None
    values = image[coords]
    output_dtype = (
            np.float64 if np.issubdtype(image.dtype, np.integer) else
            image.dtype
            )
    return values.astype(output_dtype, copy=False)


class Skeleton:
    """Object to group together all the properties of a skeleton.

    In the text below, we use the following notation:

    - N: the number of points in the pixel skeleton,
    - ndim: the dimensionality of the skeleton
    - P: the number of paths in the skeleton (also the number of links in the
      junction graph).
    - J: the number of junction nodes
    - Sd: the sum of the degrees of all the junction nodes
    - [Nt], [Np], Nr, Nc: the dimensions of the source image

    Parameters
    ----------
    skeleton_image : array
        The input skeleton (1-pixel/voxel thick skeleton, all other values 0).

    Other Parameters
    ----------------
    spacing : float or array of float, shape ``(ndim,)``
        The scale of the pixel spacing along each axis.
    source_image : array of float, same shape as `skeleton_image`
        The image that `skeleton_image` represents / summarizes / was generated
        from. This is used to produce visualizations as well as statistical
        properties of paths.
    keep_images : bool
        Whether or not to keep the original input images. These can be useful
        for visualization, but they may take up a lot of memory.
    value_is_height : bool
        Whether to consider the value of a float skeleton to be the "height"
        of the image. This can be useful e.g. when measuring lengths along
        ridges in AFM images.

    Attributes
    ----------
    graph : scipy.sparse.csr_matrix, shape (N + 1, N + 1)
        The skeleton pixel graph, where each node is a non-zero pixel in the
        input image, and each edge connects adjacent pixels. The graph is
        represented as an adjacency matrix in SciPy sparse matrix format. For
        more information see the ``scipy.sparse`` documentation as well as
        ``scipy.sparse.csgraph``. Note: pixel numbering starts at 1, so the
        shape of this matrix is ``(N + 1, N + 1)`` instead of ``(N, N)``.
    nbgraph : NBGraph
        A thin Numba wrapper around the ``csr_matrix`` format, this provides
        faster graph methods. For example, it is much faster to get a list of
        neighbors, or test for the presence of a specific edge.
    coordinates : array, shape (N, ndim)
        skeleton_pixel_id i -> coordinates[i]
        The image coordinates of each pixel in the skeleton.
        Some values in this matrix are non-sensical — you should only access
        them from node ids.
    paths : scipy.sparse.csr_matrix, shape (P, N + 1)
        A csr_matrix where element [i, j] is on if node j is in path i. This
        includes path endpoints. The number of nonzero elements is N - J + Sd.
    n_paths : int
        The number of paths, P. This is redundant information given `n_paths`,
        but it is used often enough that it is worth keeping around.
    distances : array of float, shape (P,)
        The distance of each path. Note: not initialized until `path_lengths()`
        is called on the skeleton; use path_lengths() instead
    skeleton_image : array or None
        The input skeleton image. Only present if `keep_images` is True. Set to
        False to preserve memory.
    source_image : array or None
        The image from which the skeleton was derived. Only present if
        `keep_images` is True. This is useful for visualization.
    """

    def __init__(
            self,
            skeleton_image,
            *,
            spacing=1,
            source_image=None,
            keep_images=True,
            value_is_height=False,
            ):
        graph, coords = skeleton_to_csgraph(
                skeleton_image,
                spacing=spacing,
                value_is_height=value_is_height,
                )
        self.pixel_values = _extract_values(skeleton_image, coords)
        self.graph = graph
        self.nbgraph = csr_to_nbgraph(graph, self.pixel_values)
        self.coordinates = np.transpose(coords)
        self.paths = _build_skeleton_path_graph(self.nbgraph)
        self.n_paths = self.paths.shape[0]
        self.distances = np.empty(self.n_paths, dtype=float)
        self._distances_initialized = False
        self.skeleton_image = None
        self.skeleton_shape = skeleton_image.shape
        self.skeleton_dtype = skeleton_image.dtype
        self.source_image = None
        self.degrees = np.diff(self.graph.indptr)
        self.spacing = (
                np.asarray(spacing) if not np.isscalar(spacing) else
                np.full(skeleton_image.ndim, spacing)
                )
        if keep_images:
            self.keep_images = keep_images
            self.skeleton_image = skeleton_image
            self.source_image = source_image

    @classmethod
    def from_path_graph(cls, pg: PathGraph):
        dtype = bool if pg.node_values is None else pg.node_values.dtype
        ndim = pg.node_coordinates.shape[1]
        dummy_image = np.zeros((4,) * ndim, dtype=dtype)
        dummy_image[([1, 2],) * ndim] = 1  # make a single diagonal branch
        obj = cls(dummy_image, spacing=pg.spacing, keep_images=False)
        obj.pixel_values = pg.node_values
        obj.graph = pg.adj
        obj.nbgraph = pg.nbgraph
        obj.coordinates = pg.node_coordinates
        obj.paths = pg.paths
        obj.n_paths = pg.n_paths
        obj.distances = pg.distances
        obj.skeleton_shape = np.max(obj.coordinates, axis=0) + 1
        obj.skeleton_dtype = dtype
        obj.degrees = pg.degrees
        obj.spacing = pg.spacing
        return obj

    def path(self, index):
        """Return the pixel indices of path number `index`.

        Parameters
        ----------
        index : int
            The desired path.

        Returns
        -------
        path : array of int
            The indices of the pixels belonging to the path, including
            endpoints.
        """
        # The below is equivalent to `self.paths[index].indices`, which is much
        # more elegant. However the below version is about 25x faster!
        # In [14]: %timeit mat[1].indices
        # 128 µs ± 421 ns per loop (mean ± std. dev. of 7 runs, 10000 loops each)
        # In [16]: %%timeit
        # ...: start, stop = mat.indptr[1:3]
        # ...: mat.indices[start:stop]
        # ...:
        # 5.05 µs ± 77.2 ns per loop (mean ± std. dev. of 7 runs, 100000 loops each)
        start, stop = self.paths.indptr[index:index + 2]
        return self.paths.indices[start:stop]

    def path_coordinates(self, index: int):
        """Return the image coordinates of the pixels in the path.

        Parameters
        ----------
        index : int
            The desired path.

        Returns
        -------
        path_coords : array of float
            The (image) coordinates of points on the path, including endpoints.
        """
        path_indices = self.path(index)
        return self.coordinates[path_indices]

    def path_with_data(self, index: int):
        """Return pixel indices and corresponding pixel values on a path.

        Parameters
        ----------
        index : int
            The desired path.

        Returns
        -------
        path : array of int
            The indices of pixels on the path, including endpoints.
        data : array of float
            The values of pixels on the path.
        """
        start, stop = self.paths.indptr[index:index + 2]
        return self.paths.indices[start:stop], self.paths.data[start:stop]

    def path_lengths(self):
        """Return the length of each path on the skeleton.

        Returns
        -------
        lengths : array of float
            The length of all the paths in the skeleton.
        """
        if not self._distances_initialized:
            _compute_distances(
                    self.nbgraph, self.paths.indptr, self.paths.indices,
                    self.distances
                    )
            self._distances_initialized = True
        return self.distances

    def paths_list(self):
        """List all the paths in the skeleton, including endpoints.

        Returns
        -------
        paths : list of array of int
            The list containing all the paths in the skeleton.
        """
        return [list(self.path(i)) for i in range(self.n_paths)]

    def path_label_image(self):
        """Image like self.skeleton_image with path_ids as values.

        Returns
        -------
        label_image : array of ints
            Image of the same shape as self.skeleton_image where each pixel
            has the value of its branch id + 1.
        """
        image_out = np.zeros(self.skeleton_shape, dtype=int)
        for i in range(self.n_paths):
            coords_to_wipe = self.path_coordinates(i)
            coords_idxs = tuple(np.round(coords_to_wipe).astype(int).T)
            image_out[coords_idxs] = i + 1
        return image_out

    def path_means(self):
        """Compute the mean pixel value along each path.

        Returns
        -------
        means : array of float
            The average pixel value along each path in the skeleton.
        """
        sums = np.add.reduceat(self.paths.data, self.paths.indptr[:-1])
        lengths = np.diff(self.paths.indptr)
        return sums / lengths

    def path_stdev(self):
        """Compute the standard deviation of values along each path.

        Returns
        -------
        stdevs : array of float
            The standard deviation of pixel values along each path.
        """
        data = self.paths.data
        sumsq = np.add.reduceat(data * data, self.paths.indptr[:-1])
        lengths = np.diff(self.paths.indptr)
        means = self.path_means()
        return np.sqrt(np.clip(sumsq/lengths - means*means, 0, None))

    def prune_paths(self, indices: npt.ArrayLike) -> 'Skeleton':
        """Prune nodes from the skeleton.

        Parameters
        ----------
        indices: List[int]
            List of indices to be removed.

        Retruns
        -------
        Skeleton
            A new Skeleton object pruned.
        """
        # warning: slow
        image_cp = np.copy(self.skeleton_image)
        if not np.all(np.array(indices) < self.n_paths):
            raise ValueError(
                    f'The path index {np.max(indices)} does not exist in this '
                    f'skeleton. (The highest path index is {self.n_paths}.)\n'
                    'If you obtained the index from a summary table, you '
                    'probably need to resummarize the skeleton.'
                    )
        for i in indices:
            pixel_ids_to_wipe = self.path(i)
            junctions = self.degrees[pixel_ids_to_wipe] > 2
            pixel_ids_to_wipe = pixel_ids_to_wipe[~junctions]
            coords_to_wipe = self.coordinates[pixel_ids_to_wipe]
            coords_idxs = tuple(np.round(coords_to_wipe).astype(int).T)
            image_cp[coords_idxs] = 0
        # optional cleanup:
        new_skeleton = morphology.skeletonize(image_cp.astype(bool)) * image_cp
        return Skeleton(
                new_skeleton,
                spacing=self.spacing,
                source_image=self.source_image,
                keep_images=self.keep_images
                )

    def __array__(self, dtype=None):
        """Array representation of the skeleton path labels."""
        return self.path_label_image()


def summarize(
        skel: Skeleton | PathGraph,
        *,
        value_is_height: bool = False,
        find_main_branch: bool = False,
        separator: str | None = None,
        ) -> pd.DataFrame:
    """Compute statistics for every skeleton and branch in ``skel``.

    Parameters
    ----------
    skel : skan.csr.Skeleton
        A Skeleton object.
    value_is_height : bool
        Whether to consider the value of a float skeleton to be the "height"
        of the image. This can be useful e.g. when measuring lengths along
        ridges in AFM images.
    find_main_branch : bool, optional
        Whether to compute main branches. A main branch is defined as the
        longest shortest path within a skeleton. This step is very expensive
        as it involves computing the shortest paths between all pairs of branch
        endpoints, so it is off by default.
    separator : str, optional
        Some column names are composite, e.g. ``'coord_src_0'``. The separator
        argument allows users to configure which character is used to separate
        the components. The default up to version 0.12 is '-', but will change
        to '_' in version 0.13.

    Returns
    -------
    summary : pandas.DataFrame
        A summary of the branches including branch length, mean branch value,
        branch euclidean distance, etc.
    """
    if isinstance(skel, PathGraph):
        skel = Skeleton.from_path_graph(skel)
    if separator is None:
        warnings.warn(
                "separator in column name will change to _ in version 0.13; "
                "to silence this warning, use `separator='-'` to maintain "
                "current behavior and use `separator='_'` to switch to the "
                "new default behavior.",
                np.exceptions.VisibleDeprecationWarning,
                stacklevel=2,  # make sure warning points to calling line
                )
        separator = '-'
    summary = {}
    ndim = skel.coordinates.shape[1]
    _, skeleton_ids = csgraph.connected_components(skel.graph, directed=False)
    endpoints_src = skel.paths.indices[skel.paths.indptr[:-1]]
    endpoints_dst = skel.paths.indices[skel.paths.indptr[1:] - 1]
    summary['skeleton_id'] = skeleton_ids[endpoints_src]
    summary['node_id_src'] = endpoints_src
    summary['node_id_dst'] = endpoints_dst
    summary['branch_distance'] = skel.path_lengths()
    deg_src = skel.degrees[endpoints_src]
    deg_dst = skel.degrees[endpoints_dst]
    kind = np.full(deg_src.shape, 2)  # default: junction-to-junction
    kind[(deg_src == 1) | (deg_dst == 1)] = 1  # tip-junction
    kind[(deg_src == 1) & (deg_dst == 1)] = 0  # tip-tip
    kind[endpoints_src == endpoints_dst] = 3  # cycle
    summary['branch_type'] = kind
    summary['mean_pixel_value'] = skel.path_means()
    summary['stdev_pixel_value'] = skel.path_stdev()
    for i in range(ndim):  # keep loops separate for best insertion order
        summary[f'image_coord_src_{i}'] = skel.coordinates[endpoints_src, i]
    for i in range(ndim):
        summary[f'image_coord_dst_{i}'] = skel.coordinates[endpoints_dst, i]
    coords_real_src = skel.coordinates[endpoints_src] * skel.spacing
    for i in range(ndim):
        summary[f'coord_src_{i}'] = coords_real_src[:, i]
    if value_is_height:
        values_src = skel.pixel_values[endpoints_src]
        summary[f'coord_src_{ndim}'] = values_src
        coords_real_src = np.concatenate(
                [coords_real_src, values_src[:, np.newaxis]],
                axis=1,
                )
    coords_real_dst = skel.coordinates[endpoints_dst] * skel.spacing
    for i in range(ndim):
        summary[f'coord_dst_{i}'] = coords_real_dst[:, i]
    if value_is_height:
        values_dst = skel.pixel_values[endpoints_dst]
        summary[f'coord_dst_{ndim}'] = values_dst
        coords_real_dst = np.concatenate(
                [coords_real_dst, values_dst[:, np.newaxis]],
                axis=1,
                )
    summary['euclidean_distance'] = (
            np.sqrt((coords_real_dst - coords_real_src)**2
                    @ np.ones(ndim + int(value_is_height)))
            )
    df = pd.DataFrame(summary)

    if find_main_branch:
        # define main branch as longest shortest path within a single skeleton
        df['main'] = find_main_branches(df)
    df.rename(columns=lambda s: s.replace('_', separator), inplace=True)
    return df


@numba.jit(nopython=True, nogil=True, cache=False)  # cache with Numba 1.0
def _compute_distances(graph, path_indptr, path_indices, distances):
    for i in range(len(distances)):
        start, stop = path_indptr[i:i + 2]
        path = path_indices[start:stop]
        distances[i] = _path_distance(graph, path)


@numba.jit(nopython=True, nogil=True, cache=False)  # cache with Numba 1.0
def _path_distance(graph, path):
    d = 0.
    n = len(path)
    for i in range(n - 1):
        u, v = path[i], path[i + 1]
        d += graph.edge(u, v)
    return d


def _mst_junctions(csmat):
    """Replace clustered pixels with degree > 2 by their minimum spanning tree.

    This function performs the operation in place.

    Parameters
    ----------
    csmat : NBGraph
        The input graph.
    pixel_indices : array of int
        The raveled index in the image of every pixel represented in csmat.
    spacing : float, or array-like of float, shape `len(shape)`, optional
        The spacing between pixels in the source image along each dimension.

    Returns
    -------
    final_graph : NBGraph
        The output csmat.
    """
    # make copy
    # mask out all degree < 3 entries
    # find MST
    # replace edges not in MST with zeros
    # use .eliminate_zeros() to get a new matrix
    csc_graph = csmat.tocsc()
    degrees = np.asarray(csmat.astype(bool).astype(int).sum(axis=0))
    non_junction = np.flatnonzero(degrees < 3)
    non_junction_column_start = csc_graph.indptr[non_junction]
    non_junction_column_end = csc_graph.indptr[non_junction + 1]
    for start, end in zip(non_junction_column_start, non_junction_column_end):
        csc_graph.data[start:end] = 0
    csr_graph = csc_graph.tocsr()
    non_junction_row_start = csr_graph.indptr[non_junction]
    non_junction_row_end = csr_graph.indptr[non_junction + 1]
    for start, end in zip(non_junction_row_start, non_junction_row_end):
        csr_graph.data[start:end] = 0
    csr_graph.eliminate_zeros()
    mst = csgraph.minimum_spanning_tree(csr_graph)
    non_tree_edges = csr_graph - (mst + mst.T)
    final_graph = csmat - non_tree_edges
    return final_graph


def distance_with_height(source_values, neighbor_values, distances):
    height_diff = source_values - neighbor_values
    return np.hypot(height_diff, distances)


def skeleton_to_csgraph(
        skel,
        *,
        spacing=1,
        value_is_height=False,
        ):
    """Convert a skeleton image of thin lines to a graph of neighbor pixels.

    Parameters
    ----------
    skel : array
        An input image in which every nonzero pixel is considered part of
        the skeleton, and links between pixels are determined by a full
        n-dimensional neighborhood.
    spacing : float, or array-like of float, shape `(skel.ndim,)`
        A value indicating the distance between adjacent pixels. This can
        either be a single value if the data has the same resolution along
        all axes, or it can be an array of the same shape as `skel` to
        indicate spacing along each axis.

    Other Parameters
    ----------------
    value_is_height : bool, optional
        If `True`, the pixel value at each point of the skeleton will be
        considered to be a height measurement, and this height will be
        incorporated into skeleton branch lengths. Used for analysis of
        atomic force microscopy (AFM) images.

    Returns
    -------
    graph : sparse.csr_matrix
        A graph of shape (Nnz, Nnz), where Nnz is the number of
        nonzero pixels in `skel`. The value graph[i, j] is the distance
        between adjacent pixels i and j. In a 2D image, that would be
        1 for immediately adjacent pixels and sqrt(2) for diagonally
        adjacent ones.
    pixel_coordinates : array of float
        An array of shape (Nnz, skel.ndim), mapping indices in `graph`
        to pixel coordinates in `skel`.
    """
    # ensure we have a bool image, since we later use it for bool indexing
    skel_im = skel
    skel_bool = skel.astype(bool)
    ndim = skel.ndim
    spacing = np.ones(ndim, dtype=float) * spacing

    if value_is_height:
        edge_func = distance_with_height
    else:
        edge_func = None

    graph, pixel_indices = pixel_graph(
            skel_im,
            mask=skel_bool,
            edge_function=edge_func,
            connectivity=ndim,
            spacing=spacing
            )

    graph = _mst_junctions(graph)
    pixel_coordinates = np.unravel_index(pixel_indices, skel.shape)
    return graph, pixel_coordinates


@numba.jit(nopython=True, cache=True)
def _csrget(indices, indptr, data, row, col):
    """Fast lookup of value in a scipy.sparse.csr_matrix format table.

    Parameters
    ----------
    indices, indptr, data : numpy arrays of int, int, float
        The CSR format data.
    row, col : int
        The matrix coordinates of the desired value.

    Returns
    -------
    dat: float
        The data value in the matrix.
    """
    start, end = indptr[row], indptr[row + 1]
    for i in range(start, end):
        if indices[i] == col:
            return data[i]
    return 0.


@numba.jit(nopython=True, cache=True)
def _csrset(indices, indptr, data, row, col, value):
    """Fast lookup and set of value in a scipy.sparse.csr_matrix format table.

    Parameters
    ----------
    indices, indptr, data : numpy arrays of int, int, float
        The CSR format data.
    row, col : int
        The matrix coordinates of the desired value.
    value : dtype
        The value to set in the matrix.

    Notes
    -----
    This function only sets values that already existed in the matrix.

    Returns
    -------
    success: bool
        Whether the data value was successfully written to the matrix.
    """
    start, end = indptr[row], indptr[row + 1]
    for i in range(start, end):
        if indices[i] == col:
            data[i] = value
            return True
    return False


def submatrix(M, idxs):
    """Return the outer-index product submatrix, `M[idxs, :][:, idxs]`.

    Parameters
    ----------
    M : scipy.sparse.spmatrix
        Input (square) matrix
    idxs : array of int
        The indices to subset. No index in `idxs` should exceed the
        number of rows of `M`.

    Returns
    -------
    Msub : scipy.sparse.spmatrix
        The subsetted matrix.

    Examples
    --------
    >>> Md = np.arange(16).reshape((4, 4))
    >>> M = sparse.csr_matrix(Md)
    >>> print(submatrix(M, [0, 2]).toarray())
    [[ 0  2]
     [ 8 10]]
    """
    Msub = M[idxs, :][:, idxs]
    return Msub


def make_degree_image(skeleton_image):
    """Create a array showing the degree of connectivity of each pixel.

    Parameters
    ----------
    skeleton_image : array
        An input image in which every nonzero pixel is considered part of
        the skeleton, and links between pixels are determined by a full
        n-dimensional neighborhood.

    Returns
    -------
    degree_image : array of int, same shape as skeleton_image
        An image containing the degree of connectivity of each pixel in the
        skeleton to neighboring pixels.
    """
    bool_skeleton = skeleton_image.astype(bool)
    degree_kernel = np.ones((3,) * bool_skeleton.ndim)
    degree_kernel[(1,) * bool_skeleton.ndim] = 0  # remove centre pixel
    if isinstance(bool_skeleton, np.ndarray):
        degree_image = ndi.convolve(
                bool_skeleton.astype(int),
                degree_kernel,
                mode='constant',
                ) * bool_skeleton
    # use dask image for any array other than a numpy array (which isn't
    # supported yet anyway)
    else:
        import dask.array as da
        from dask_image.ndfilters import convolve as dask_convolve

        if isinstance(bool_skeleton, da.Array):
            degree_image = bool_skeleton * dask_convolve(
                    bool_skeleton.astype(int), degree_kernel, mode='constant'
                    )
    return degree_image


def _simplify_graph(skel):
    """Iterative removal of all nodes of degree 2 while reconnecting their
    edges.

    Parameters
    ----------
    skel : skan.csr.Skeleton
        A Skeleton object containing graph to be simplified.

    Returns
    -------
    simp_csgraph : scipy.sparse.csr_matrix
        A sparse adjacency matrix of the simplified graph.
    reduced_nodes : tuple of int
        The index nodes of original graph in simplified graph.
    """
    if np.sum(skel.degrees > 2) == 0:  # no junctions
        # don't reduce
        return skel.graph, np.arange(skel.graph.shape[0])

    summary = summarize(skel, separator='_')
    src = np.asarray(summary['node_id_src'])
    dst = np.asarray(summary['node_id_dst'])
    distance = np.asarray(summary['branch_distance'])

    # Workaround for https://github.com/scikit-image/scikit-image/issues/6378
    # Remove when https://github.com/scikit-image/scikit-image/pull/7535 is
    # included in a release.
    # Writing into these arrays would be Bad, but we have no intention of doing
    # so, it's just that Cython won't take a read-only array if you haven't
    # explicitly declared that your function won't attempt to write to it.
    src.flags.writeable = True
    dst.flags.writeable = True

    # to reduce the size of simplified graph
    nodes = np.unique(np.append(src, dst))
    n_nodes = len(nodes)
    nodes_sequential = np.arange(n_nodes)

    fw_map = ArrayMap(nodes, nodes_sequential)
    inv_map = ArrayMap(nodes_sequential, nodes)

    src_relab, dst_relab = fw_map[src], fw_map[dst]

    edges = sparse.coo_matrix((distance, (src_relab, dst_relab)),
                              shape=(n_nodes, n_nodes))
    dir_csgraph = edges.tocsr()
    simp_csgraph = dir_csgraph + dir_csgraph.T  # make undirected

    reduced_nodes = inv_map[np.arange(simp_csgraph.shape[0])]

    return simp_csgraph, reduced_nodes


def _fast_graph_center_idx(skel):
    """Accelerated graph center finding using simplified graph.

    Parameters
    ----------
    skel : skan.csr.Skeleton
        A Skeleton object containing graph whose center is to be found.

    Returns
    -------
    original_center_idx : int
        The index of central node of graph.
    """
    simp_csgraph, reduced_nodes = _simplify_graph(skel)
    simp_center_idx, _ = central_pixel(simp_csgraph)
    original_center_idx = reduced_nodes[simp_center_idx]

    return original_center_idx


def _normalize_shells(shells, *, center, skeleton_coordinates, spacing):
    """Normalize shells from any format allowed by `sholl_analysis` to radii.

    Parameters
    ----------
    shells : int or sequence of floats, or None
        If an int, it is used as number of evenly spaced concentric shells. If
        an array of floats, it is used directly as the different shell radii in
        real world units. If None, the number of evenly spaced concentric
        shells is automatically calculated.
    center : (D,) array of float
        The scaled coordinates of the center point for Sholl analysis.
    skeleton_coordinates : (N, D) array of float
        The scaled coordinates of skeleton pixels. Used when shells is None or
        int.
    spacing : (D,) array of float
        The pixel/voxel spacing of the skeleton data.

    Returns
    -------
    radii : array of float
        The computed and normalized shell radii.
    """
    if isinstance(shells, (list, tuple, np.ndarray)):
        shell_radii = np.asarray(shells)
    else:  # shells is int, number of shells, or None
        # Find max euclidean distance from center to all nodes
        distances = np.linalg.norm(skeleton_coordinates - center, axis=1)
        start_radius = 0
        end_radius = np.max(distances)  # largest possible radius
        if shells is None:
            stepsize = np.linalg.norm(spacing)
        else:  # scalar
            stepsize = (end_radius-start_radius) / shells
        epsilon = np.finfo(np.float32).eps
        shell_radii = np.arange(start_radius, end_radius + epsilon, stepsize)
    if (sp := np.linalg.norm(spacing)) > (sh := np.min(np.diff(shell_radii))):
        warnings.warn(
                'This implementation of Sholl analysis may not be accurate if '
                'the spacing between shells is smaller than the (diagonal) '
                f'voxel spacing. The given voxel spacing is {sp}, and the '
                f'smallest shell spacing is {sh}.',
                stacklevel=2
                )
    return shell_radii


def sholl_analysis(skeleton, center=None, shells=None):
    """Sholl Analysis for Skeleton object.

    Parameters
    ----------
    skeleton : skan.csr.Skeleton
        A Skeleton object.
    center : array-like of float or None, optional
        Scaled coordinates of a point on the skeleton to use as the center
        from which the concentric shells are computed. If None, the
        geodesic center of skeleton is chosen.
    shells : int or array of floats or None, optional
        If an int, it is used as number of evenly spaced concentric shells. If
        an array of floats, it is used directly as the different shell radii in
        real world units. If None, the number of evenly spaced concentric
        shells is automatically calculated.

    Returns
    -------
    center : array of float
        The scaled coordinates in real world units of the center of the shells.
        (This might be provided as input, but it might also have been computed
        within this function, and that computation is expensive, so we return
        it just in case.)
    shell_radii : array of float
        Radii in real world units for concentric shells used for analysis.
    intersection_counts : array of int
        Number of intersections for corresponding shell radii.
    """
    if center is None:
        # By default, find the geodesic center of the graph
        center_idx = _fast_graph_center_idx(skeleton)
        center = skeleton.coordinates[center_idx] * skeleton.spacing
    else:
        center = np.asarray(center)

    scaled_coords = skeleton.coordinates * skeleton.spacing

    shell_radii = _normalize_shells(
            shells,
            center=center,
            skeleton_coordinates=scaled_coords,
            spacing=skeleton.spacing,
            )

    edges = skeleton.graph.tocoo()
    coords0 = scaled_coords[edges.row]
    coords1 = scaled_coords[edges.col]
    d0 = distance_matrix(coords0, [center]).ravel()
    d1 = distance_matrix(coords1, [center]).ravel()
    bins0 = np.digitize(d0, shell_radii)
    bins1 = np.digitize(d1, shell_radii)
    crossings = bins0 != bins1
    shells = np.minimum(bins0[crossings], bins1[crossings])
    # we divide by 2 because the graph is undirected, so each edge appears
    # twice in the matrix
    intersection_counts = np.bincount(shells, minlength=len(shell_radii)) // 2

    return center, shell_radii, intersection_counts


def skeleton_to_nx(
        skeleton: Skeleton,
        summary: pd.DataFrame | None = None,
        ) -> nx.MultiGraph:
    """Convert a Skeleton object to a networkx Graph.

    Parameters
    ----------
    skeleton : Skeleton
        The skeleton to convert.
    summary : pd.DataFrame | None
        The summary statistics of the skeleton. Each row in the summary table
        is an edge in the networkx graph. It is not necessary to pass this in
        because it can be computed from the input skeleton, but if it is
        already computed, it will speed up this function.

    Returns
    -------
    g : nx.MultiGraph
        A graph where each node is a junction or endpoint in the skeleton, and
        each edge is a path.
    """
    if summary is None:
        summary = summarize(skeleton, separator='_')
    # Ensure underscores in column names
    csum = summary.rename(columns=lambda s: s.replace('-', '_'))
    g = nx.MultiGraph(
            shape=skeleton.skeleton_shape, dtype=skeleton.skeleton_dtype
            )
    for row in csum.itertuples(name='Edge'):
        index = row.Index
        i = row.node_id_src
        j = row.node_id_dst
        indices, values = skeleton.path_with_data(index)
        # Nodes are added if they don't exist so only need to add edges
        g.add_edge(
                i, j, **{
                        'path': skeleton.path_coordinates(index),
                        'indices': indices,
                        'values': values,
                        }
                )
    return g


def nx_to_skeleton(g: nx.Graph | nx.MultiGraph) -> Skeleton:
    """Convert a Networkx Graph to a Skeleton object.

    The NetworkX Graph should have the following graph properties:
    - 'shape': the shape of the image from which the graph was created.
    - 'dtype': the dtype of the image from which the graph was created.

    and the following edge properties:
    - 'path': an (N, Ndim) array of coordinates in the image of the pixels
      traced by this edge.
    - 'values': an (N,) array of values for the pixels traced by this edge.

    The `skeleton_to_nx` function can produce such a graph from a `Skeleton`
    object.

    Parameters
    ----------
    g: nx.Graph
        Networkx graph to convert to Skeleton.

    Returns
    -------
    Skeleton
        Skeleton object corresponding to the input graph.

    Notes
    -----
    Currently, this method uses the naive, brute-force approach of regenerating
    the entire image from the graph, and then generating the Skeleton from the
    image. It is probably low-hanging fruit to instead generate the Skeleton
    compressed-sparse-row (CSR) data directly.
    """
    image = np.zeros(g.graph['shape'], dtype=g.graph['dtype'])
    all_coords = np.concatenate([path for _, _, path in g.edges.data('path')],
                                axis=0)
    all_values = np.concatenate([
            values for _, _, values in g.edges.data('values')
            ],
                                axis=0)

    image[tuple(all_coords.T)] = all_values

    return Skeleton(image)


def _merge_paths(p1: npt.NDArray, p2: npt.NDArray):
    """Join two paths together that have a common endpoint."""
    return np.concatenate([p1[:-1], p2], axis=0)


def _merge_edges(g: nx.Graph, e1: tuple[int], e2: tuple[int]):
    middle_node = set(e1) & set(e2)
    new_edge = sorted(
            (set(e1) | set(e2)) - {middle_node},
            key=lambda i: i in e2,
            )
    d1 = g.edges[e1]
    d2 = g.edges[e2]
    p1 = d1['path'] if e1[1] == middle_node else d1['path'][::-1]
    p2 = d2['path'] if e2[0] == middle_node else d2['path'][::-1]
    n1 = len(d1['path'])
    n2 = len(d2['path'])
    new_edge_values = {
            'skeleton_id':
                    g.edges[e1]['skeleton_id'],
            'node_id_src':
                    new_edge[0],
            'node_id_dst':
                    new_edge[1],
            'branch_distance':
                    d1['branch_distance'] + d2['branch_distance'],
            'branch_type':
                    min(d1['branch_type'], d2['branch_type']),
            'mean_pixel_value': (
                    n1 * d1['mean_pixel_value'] + n2 * d2['mean_pixel_value']
                    ) / (n1+n2),
            'stdev_pixel_value':
                    np.sqrt((
                            d1['stdev_pixel_value']**2 *
                            (n1-1) + d2['stdev_pixel_value']**2 * (n2-1)
                            ) / (n1+n2-1)),
            'path':
                    _merge_paths(p1, p2),
            }
    g.add_edge(new_edge[0], new_edge[1], **new_edge_values)
    g.remove_node(middle_node)


def _remove_simple_path_nodes(g):
    """Remove any nodes of degree 2 by merging their incident edges."""
    to_remove = [n for n in g.nodes if g.degree(n) == 2]
    for u in to_remove:
        v, w = g[u].keys()
        _merge_edges(g, (u, v), (u, w))



import numpy as np
import matplotlib.pyplot as plt
from matplotlib import collections
from matplotlib.patches import Circle
import networkx as nx
from skimage import img_as_float, morphology
from skimage.color import gray2rgb


def _normalise_image(image, *, image_cmap=None):
    image = img_as_float(image)
    if image.ndim == 2:
        if image_cmap is None:
            image = gray2rgb(image)
        else:
            image = plt.get_cmap(image_cmap)(image)[..., :3]
    return image


def pixel_perfect_figsize(image, dpi=80):
    """Return the Matplotlib figure size tuple (w, h) for given image and dpi.

    Parameters
    ----------
    image : array, shape (M, N[, 3])
        The image to be plotted.
    dpi : int, optional
        The desired figure dpi.

    Returns
    -------
    figsize : tuple of float
        The desired figure size.

    Examples
    --------
    >>> image = np.empty((768, 1024))
    >>> pixel_perfect_figsize(image)
    (12.8, 9.6)
    """
    hpix, wpix = image.shape[:2]
    return wpix / dpi, hpix / dpi


def overlay_skeleton_2d(
        image,
        skeleton,
        *,
        image_cmap=None,
        color=(1, 0, 0),
        alpha=1,
        dilate=0,
        axes=None
        ):
    """Overlay the skeleton pixels on the input image.

    Parameters
    ----------
    image : array, shape (M, N[, 3])
        The input image. Can be grayscale or RGB.
    skeleton : array, shape (M, N)
        The input 1-pixel-wide skeleton.

    Other Parameters
    ----------------
    image_cmap : matplotlib colormap name or object, optional
        If the input image is grayscale, colormap it with this colormap.
        The default is grayscale.
    color : tuple of float in [0, 1], optional
        The RGB color for the skeleton pixels.
    alpha : float, optional
        Blend the skeleton pixels with the given alpha.
    dilate : int, optional
        Dilate the skeleton by this amount. This is useful when rendering
        large images where aliasing may cause some pixels of the skeleton
        not to be drawn.
    axes : matplotlib Axes
        The Axes on which to plot the image. If None, new ones are created.

    Returns
    -------
    axes : matplotlib Axes
        The Axis on which the image is drawn.
    """
    image = _normalise_image(image, image_cmap=image_cmap)
    skeleton = skeleton.astype(bool)
    if dilate > 0:
        selem = morphology.disk(dilate)
        skeleton = morphology.binary_dilation(skeleton, selem)
    if axes is None:
        fig, axes = plt.subplots()
    image[skeleton] = alpha * np.array(color) + (1-alpha) * image[skeleton]
    axes.imshow(image)
    axes.axis('off')
    return axes


def overlay_euclidean_skeleton_2d(
        image,
        stats,
        *,
        image_cmap=None,
        skeleton_color_source='branch_type',
        skeleton_colormap='viridis',
        axes=None
        ):
    """Plot the image, and overlay the straight-line skeleton over it.

    Parameters
    ----------
    image : array, shape (M, N)
        The input image.
    stats : array, shape (M, N)
        Skeleton statistics.

    Other Parameters
    ----------------
    image_cmap : matplotlib colormap name or object, optional
        The colormap to use for the input image. Defaults to grayscale.
    skeleton_color_source : string, optional
        The name of the column to use for the skeleton edge color. See the
        output of `skan.summarise` for valid choices. Most common choices
        would be:

        - skeleton-id: each individual skeleton (connected component) will
          have a different colour.
        - branch_type: each branch type (tip-tip, tip-junction,
          junction-junction, path-path). This is the default.
        - branch-distance: the curved length of the skeleton branch.
        - euclidean-distance: the straight-line length of the skeleton branch.

    skeleton_colormap : matplotlib colormap name or object, optional
        The colormap for the skeleton values.
    axes : matplotlib Axes object, optional
        An Axes object on which to draw. If `None`, a new one is created.

    Returns
    -------
    axes : matplotlib Axes object
        The Axes on which the plot is drawn.
    """
    image = _normalise_image(image, image_cmap=image_cmap)
    summary = stats
    # transforming from row, col to x, y
    coords_cols = (['image_coord_src_%i' % i for i in [1, 0]]
                   + ['image_coord_dst_%i' % i for i in [1, 0]])
    coords = summary[coords_cols].values.reshape((-1, 2, 2))
    if axes is None:
        fig, axes = plt.subplots()
    axes.imshow(image)
    axes.axis('off')
    color_values = summary[skeleton_color_source]
    cmap = plt.get_cmap(
            skeleton_colormap, min(len(np.unique(color_values)), 256)
            )
    colormapped = cmap((color_values - np.min(color_values)) /
                       (np.max(color_values) - np.min(color_values)))
    linecoll = collections.LineCollection(coords, colors=colormapped)
    axes.add_collection(linecoll)
    return axes


def overlay_skeleton_2d_class(
        skeleton,
        *,
        image_cmap='gray',
        skeleton_color_source='path_means',
        skeleton_colormap='viridis',
        vmin=None,
        vmax=None,
        axes=None
        ):
    """Plot the image, and overlay the skeleton over it.

    Parameters
    ----------
    skeleton : skan.Skeleton object
        The input skeleton, which contains both the skeleton and the source
        image.

    Other Parameters
    ----------------
    image_cmap : matplotlib colormap name or object, optional
        The colormap to use for the input image. Defaults to grayscale.
    skeleton_color_source : string or callable, optional
        The name of the method to use for the skeleton edge color. See the
        documentation of `skan.Skeleton` for valid choices. Most common choices
        would be:

        - path_means: the mean value of the skeleton along each path.
        - path_lengths: the length of each path.
        - path_stdev: the standard deviation of pixel values along the path.

        Alternatively, a callable can be provided that takes as input a
        Skeleton object and outputs a list of floating point values of the same
        length as the number of paths.

    skeleton_colormap : matplotlib colormap name or object, optional
        The colormap for the skeleton values.
    vmin, vmax : float, optional
        The minimum and maximum values for the colormap. Use this to pin the
        colormapped values to a certain range.
    axes : matplotlib Axes object, optional
        An Axes object on which to draw. If `None`, a new one is created.

    Returns
    -------
    axes : matplotlib Axes object
        The Axes on which the plot is drawn.
    mappable : matplotlib ScalarMappable object
        The mappable values corresponding to the line colors. This can be used
        to create a colorbar for the plot.
    """
    if axes is None:
        fig, axes = plt.subplots()
    if skeleton.source_image is not None:
        axes.imshow(skeleton.source_image, cmap=image_cmap)
    if callable(skeleton_color_source):
        values = skeleton_color_source(skeleton)
    elif hasattr(skeleton, skeleton_color_source):
        values = getattr(skeleton, skeleton_color_source)()
    else:
        raise ValueError(
                'Unknown skeleton color source: %s. Provide an '
                'attribute of skan.csr.Skeleton or a callable.'
                % skeleton_color_source
                )
    cmap = plt.get_cmap(skeleton_colormap, min(len(np.unique(values)), 256))
    if vmin is None:
        vmin = np.min(values)
    if vmax is None:
        vmax = np.max(values)
    mapping_values = (values-vmin) / (vmax-vmin)
    mappable = plt.cm.ScalarMappable(plt.Normalize(vmin, vmax), cmap)
    mappable._A = mapping_values
    colors = cmap(mapping_values)
    coordinates = [
            skeleton.path_coordinates(i)[:, ::-1]
            for i in range(skeleton.n_paths)
            ]
    linecoll = collections.LineCollection(coordinates, colors=colors)
    axes.add_collection(linecoll)
    return axes, mappable


def sholl_shells(center, radii, *, axes=None, **kwargs):
    """Draw concentric circles around a center point.

    Parameters
    ----------
    center : array of float, shape (2,)
        The center of the circles. This should be in NumPy-style row/column
        coordinates.
    radii : array of float, shape (N,)
        The radii of the concentric circles.
    axes : matplotlib Axes, optional
        The axes on which to draw the circles. If None, create a new instance.

    Returns
    -------
    axes : matplotlib Axes
        The axes on which the circles were drawn
    patches : list of matplotlib Patches
        The patch objects that were drawn.

    Notes
    -----
    Additional keyword arguments are passed directly to the
    `matplotlib.patches.Circle` call. Valid keywords include ``edgecolor``,
    ``linestyle``, and `linewidth``. See matplotlib documentation for details.
    """
    row, col = center
    color = kwargs.pop('edgecolor', 'cornflowerblue')
    circles = [
            Circle((col, row), radius=r, fill=False, edgecolor=color, **kwargs)
            for r in radii
            ]
    if axes is None:
        _, axes = plt.subplots()
    for c in circles:
        axes.add_patch(c)
    return axes, circles


def pipeline_plot(
        image,
        thresholded,
        skeleton,
        stats,
        *,
        figure=None,
        axes=None,
        figsize=(9, 9)
        ):
    """Draw the image, the thresholded version, and its skeleton.

    Parameters
    ----------
    image : array, shape (M, N, ...[, 3])
        Input image, conformant with scikit-image data type
        specification [1]_.
    thresholded : array, same shape as image
        Binarized version of the input image.
    skeleton : array, same shape as image
        Skeletonized version of the input image.
    stats : pandas DataFrame
        Skeleton statistics from the input image/skeleton.

    Other Parameters
    ----------------
    figure : matplotlib Figure, optional
        If given, where to make the plots.
    axes : array of matplotlib Axes, optional
        If given, use these axes to draw the plots. Should have len 4.
    figsize : 2-tuple of float, optional
        The width and height of the figure.
    smooth_method : {'Gaussian', 'TV', 'NL'}, optional
        Which denoising method to use on the image.

    Returns
    -------
    fig : matplotlib Figure
        The Figure containing all the plots
    axes : array of matplotlib Axes
        The four axes containing the drawn images.

    References
    ----------
    .. [1] http://scikit-image.org/docs/dev/user_guide/data_types.html
    """
    if figure is None:
        fig, axes = plt.subplots(
                2, 2, figsize=figsize, sharex=True, sharey=True
                )
        axes = np.ravel(axes)
    else:
        fig = figure
        if axes is None:
            ax0 = fig.add_subplot(2, 2, 1)
            axes = [ax0] + [
                    fig.add_subplot(2, 2, i, sharex=ax0, sharey=ax0)
                    for i in range(2, 5)
                    ]

    axes = np.ravel(axes)
    axes[0].imshow(image, cmap='gray')
    axes[0].axis('off')

    axes[1].imshow(thresholded, cmap='gray')
    axes[1].axis('off')

    overlay_skeleton_2d(image, skeleton, axes=axes[2])

    overlay_euclidean_skeleton_2d(image, stats, axes=axes[3])

    fig.subplots_adjust(0, 0, 1, 1, 0, 0)

    return fig, axes


def _clean_positions_dict(d, g):
    for k in list(d.keys()):
        if k not in g:
            del d[k]
        elif g.degree(k) == 0:
            g.remove_node(k)


def overlay_skeleton_networkx(
        csr_graph, coordinates, *, axis=None, image=None, cmap=None, **kwargs
        ):
    """Draw the skeleton as a NetworkX graph, optionally overlaid on an image.

    Due to the size of NetworkX drawing elements, this is only recommended
    for very small skeletons.

    Parameters
    ----------
    csr_graph : SciPy Sparse matrix
        The skeleton graph in SciPy CSR format.
    coordinates : array, shape (N_points, 2)
        The coordinates of each point in the skeleton. ``coordinates.shape[0]``
        should be equal to ``csr_graph.shape[0]``.

    Other Parameters
    ----------------
    axis : Matplotlib Axes object, optional
        The Axes on which to plot the data. If None, a new figure and axes will
        be created.
    image : array, shape (M, N[, 3])
        An image on which to overlay the skeleton. ``image.shape`` should be
        greater than ``np.max(coordinates, axis=0)``.
    **kwargs : keyword arguments
        Arguments passed on to `nx.draw_networkx`. Particularly useful ones
        include ``node_size=`` and ``font_size=``.
    """
    if axis is None:
        _, axis = plt.subplots()
    if image is not None:
        cmap = cmap or 'gray'
        axis.imshow(image, cmap=cmap)
    gnx = nx.from_scipy_sparse_array(csr_graph)
    # Note: we invert the positions because Matplotlib uses x/y for
    # scatterplot, but the coordinates are row/column NumPy indexing
    positions = dict(zip(range(coordinates.shape[0]), coordinates[:, ::-1]))
    _clean_positions_dict(positions, gnx)  # remove nodes not in Graph
    nx.draw_networkx(gnx, pos=positions, ax=axis, **kwargs)
    return axis



import sys
from time import process_time
import numpy as np
from numpy.testing import assert_allclose
import pytest
from skan.csr import Skeleton, summarize

from skan._testdata import (
        tinycycle, skeleton1, skeleton2, skeleton3d, skeleton4, junction_first
        )


def test_tiny_cycle():
    skeleton = Skeleton(tinycycle)
    assert skeleton.paths.shape == (1, 5)


def test_skeleton1_topo():
    skeleton = Skeleton(skeleton1)
    assert skeleton.paths.shape == (4, 21)
    paths_list = skeleton.paths_list()
    reference_paths = [[7, 5, 0, 1, 2, 3, 4, 6, 10, 9, 12], [7, 8, 12],
                       [7, 11, 13], [12, 14, 15, 16]]
    d0 = 1 + np.sqrt(2)
    reference_distances = [5 * d0, d0, d0, 1 + d0]
    for path in reference_paths:
        assert path in paths_list or path[::-1] in paths_list
    assert_allclose(
            sorted(skeleton.path_lengths()), sorted(reference_distances)
            )


def test_skeleton1_float():
    image = np.zeros(skeleton1.shape, dtype=float)
    image[skeleton1] = 1 + np.random.random(np.sum(skeleton1))
    skeleton = Skeleton(image)
    path, data = skeleton.path_with_data(0)
    assert 1.0 < np.mean(data) < 2.0


def test_skeleton_coordinates():
    skeleton = Skeleton(skeleton1)
    last_path_coordinates = skeleton.path_coordinates(3)
    assert_allclose(last_path_coordinates, [[3, 3], [4, 4], [4, 5], [4, 6]])


@pytest.mark.skipif(
        sys.platform.startswith('win'),
        reason='caching not working on Windows CI for mysterious reasons',
        )
def test_path_length_caching():
    skeleton = Skeleton(skeleton3d)
    t0 = process_time()
    distances = skeleton.path_lengths()
    t1 = process_time()
    distances2 = skeleton.path_lengths()
    t2 = process_time()
    assert t2 - t1 < t1 - t0
    assert np.all((distances >= np.sqrt(2))
                  & (distances < 5.01 + 5 * np.sqrt(2)))


def test_tip_junction_edges():
    skeleton = Skeleton(skeleton4)
    reference_paths = [[1, 0], [1, 2, 3, 4], [1, 5, 6]]
    paths_list = skeleton.paths_list()
    for path in reference_paths:
        assert path in paths_list or path[::-1] in paths_list


def test_path_stdev():
    image = np.zeros(skeleton1.shape, dtype=float)
    image[skeleton1] = 1 + np.random.random(np.sum(skeleton1))
    skeleton = Skeleton(image)
    # longest_path should be 0, but could change.
    longest_path = np.argmax(skeleton.path_lengths())
    dev = skeleton.path_stdev()[longest_path]
    assert 0.09 < dev < 0.44  # chance is < 1/10K that this will fail

    # second check: first principles.
    skeleton2 = Skeleton(image**2)
    # (Var = StDev**2 = E(X**2) - (E(X))**2)
    assert_allclose(
            skeleton.path_stdev()**2,
            skeleton2.path_means() - skeleton.path_means()**2
            )


def test_junction_first():
    """Ensure no self-edges exist in multi-pixel junctions.

    Before commit 64047622, the skeleton class would include self-edges
    within junctions in its paths list, but only when the junction was visited
    before any of its adjacent branches. This turns out to be tricky to achieve
    but not impossible in 2D.
    """
    assert [1, 1] not in Skeleton(junction_first).paths_list()


def test_skeleton_summarize():
    image = np.zeros(skeleton2.shape, dtype=float)
    image[skeleton2] = 1 + np.random.random(np.sum(skeleton2))
    skeleton = Skeleton(image)
    summary = summarize(skeleton, separator='_')
    assert set(summary['skeleton_id']) == {0, 1}
    assert (
            np.all(summary['mean_pixel_value'] < 2)
            and np.all(summary['mean_pixel_value'] > 1)
            )


def test_skeleton_label_image_strict():
    """Test that the skeleton pixel labels match the branch IDs.

    This does pixel-wise pairing of the label image with the expected label
    image.  There should be the same number of unique pairs as there are unique
    labels in the expected label image. This assumes that the branches are
    displayed correctly, but does not assert an order to the numbering of the
    branches. Junction pixel IDs are ambiguous and so are not checked. It is an
    implementation detail that currently, the latest branch ID touching that
    junction "wins".
    """
    skeleton = Skeleton(skeleton4)
    label_image = np.asarray(skeleton)
    expected = np.array([
            [1, 0, 0, 0, 0],
            [0, 3, 2, 2, 2],
            [0, 3, 0, 0, 0],
            [0, 3, 0, 0, 0],
            ])
    non_junction_pixels = (skeleton.degrees < 3)
    non_junction_coords = tuple(
            np.transpose(skeleton.coordinates[non_junction_pixels])
            )
    paired_values = np.stack(
            (expected[non_junction_coords], label_image[non_junction_coords]),
            axis=-1
            ).reshape((-1, 2))
    unique_pairs = np.unique(paired_values, axis=0)
    # check that the number of pairs is equal to the number of values
    expected_label_values = np.unique(expected[expected > 0])
    assert len(expected_label_values) == len(unique_pairs)
    # check that all nonzero pixels are preserved
    np.testing.assert_array_equal(expected != 0, label_image != 0)


def test_skeleton_label_image():
    """Simple test that the skeleton label image covers the same
    pixels as the expected label image.
    """
    skeleton = Skeleton(skeleton4)
    label_image = np.asarray(skeleton)
    expected = np.array([
            [1, 0, 0, 0, 0],
            [0, 1, 2, 2, 2],
            [0, 3, 0, 0, 0],
            [0, 3, 0, 0, 0],
            ])

    np.testing.assert_array_equal(
            label_image.astype(bool), expected.astype(bool)
            )



import pytest

from skan._testdata import skeleton0
from skan import Skeleton


@pytest.mark.parametrize('branch_num', [0, 1, 2])
def test_pruning_comprehensive(branch_num):
    skeleton = Skeleton(skeleton0)
    pruned = skeleton.prune_paths([branch_num])
    print(pruned.skeleton_image.astype(int))
    assert pruned.n_paths == 1



from __future__ import annotations
import sys

from collections import defaultdict
from itertools import product
from typing import Any

import numpy as np
import numpy.typing as npt
from numpy.testing import assert_equal, assert_almost_equal
import pandas as pd
import pytest
import scipy
from scipy import ndimage as ndi
from skimage import data
from skimage.draw import line
from skimage.morphology import skeletonize

from skan import csr
from skan._testdata import (
        tinycycle,
        tinyline,
        skeleton0,
        skeleton1,
        skeleton2,
        skeleton3d,
        topograph1d,
        skeleton4,
        skeletonlabel,
        skeleton_loop1,
        skeleton_loop2,
        skeleton_linear1,
        skeleton_linear2,
        skeleton_linear3,
        nx_graph,
        nx_graph_edges,
        )


def _old_branch_statistics(
        skeleton_image, *, spacing=1, value_is_height=False
        ):
    skel = csr.Skeleton(
            skeleton_image, spacing=spacing, value_is_height=value_is_height
            )
    summary = csr.summarize(
            skel, value_is_height=value_is_height, separator='_'
            )
    columns = ['node_id_src', 'node_id_dst', 'branch_distance', 'branch_type']
    return summary[columns].to_numpy()


def test_tiny_cycle():
    g, idxs = csr.skeleton_to_csgraph(tinycycle)
    expected_indptr = [0, 2, 4, 6, 8]
    expected_indices = [1, 2, 0, 3, 0, 3, 1, 2]
    expected_data = np.sqrt(2)

    assert_equal(g.indptr, expected_indptr)
    assert_equal(g.indices, expected_indices)
    assert_almost_equal(g.data, expected_data)

    assert_equal(np.ravel_multi_index(idxs, tinycycle.shape), [1, 3, 5, 7])


def test_skeleton1_stats():
    stats = _old_branch_statistics(skeleton1)
    assert_equal(stats.shape, (4, 4))
    keys = map(tuple, np.sort(stats[:, :2].astype(int), axis=1))
    dists = stats[:, 2]
    types = stats[:, 3].astype(int)
    ids2dist = defaultdict(list)
    for key, dist in zip(keys, dists):
        ids2dist[key].append(dist)
    assert (7, 12) in ids2dist
    d0, d1 = sorted(ids2dist[(7, 12)])
    assert_almost_equal(d0, 1 + np.sqrt(2))
    assert_almost_equal(d1, 5 * d0)
    assert_equal(np.bincount(types), [0, 2, 2])
    assert_almost_equal(np.unique(dists), [d0, 2 + np.sqrt(2), d1])


def test_2skeletons():
    df = csr.summarize(csr.Skeleton(skeleton2), separator='_')
    assert_almost_equal(np.unique(df['euclidean_distance']), np.sqrt([5, 10]))
    assert_equal(np.unique(df['skeleton_id']), [0, 1])
    assert_equal(np.bincount(df['branch_type']), [0, 4, 4])


def test_summarize_spacing():
    df = csr.summarize(csr.Skeleton(skeleton2), separator='_')
    df2 = csr.summarize(csr.Skeleton(skeleton2, spacing=2), separator='_')
    assert_equal(np.array(df['node_id_src']), np.array(df2['node_id_src']))
    assert_almost_equal(
            np.array(df2['euclidean_distance']),
            np.array(2 * df['euclidean_distance'])
            )
    assert_almost_equal(
            np.array(df2['branch_distance']),
            np.array(2 * df['branch_distance'])
            )


def test_line():
    g, idxs = csr.skeleton_to_csgraph(tinyline)
    assert_equal(np.ravel(idxs), [1, 2, 3])
    assert_equal(g.shape, (3, 3))
    # source, dest, length, type
    assert_equal(_old_branch_statistics(tinyline), [[0, 2, 2, 0]])


def test_cycle_stats():
    stats = _old_branch_statistics(tinycycle)
    # source, dest, length, type
    assert_almost_equal(stats, [[0, 0, 4 * np.sqrt(2), 3]])


def test_3d_spacing():
    stats = _old_branch_statistics(skeleton3d, spacing=[5, 1, 1])
    assert_equal(stats.shape, (7, 4))
    assert_almost_equal(stats[0], [0, 10, 2 * np.sqrt(5**2 + 1 + 1), 1])
    # source, dest, length, type
    # test only junction-tip segments
    assert_equal(np.unique(stats[:, 3].astype(int)), [1, 2, 3])


def test_topograph():
    stats = _old_branch_statistics(topograph1d, value_is_height=True)
    assert stats.shape == (1, 4)
    assert_almost_equal(stats[0], [0, 2, 2 * np.sqrt(2), 0])


def test_topograph_summary():
    stats = csr.summarize(
            csr.Skeleton(topograph1d, spacing=2.5, value_is_height=True),
            value_is_height=True,
            separator='_',
            )
    assert stats.loc[0, 'euclidean_distance'] == 5.0
    columns = ['coord_src_0', 'coord_src_1', 'coord_dst_0', 'coord_dst_1']
    assert_almost_equal(sorted(stats.loc[0, columns]), [0, 3, 3, 5])


def test_junction_multiplicity():
    """Test correct distances when a junction has more than one pixel."""
    g, _ = csr.skeleton_to_csgraph(skeleton0)
    assert_equal(g.data, 1.0)
    assert_equal(g[2, 5], 0.0)


def test_multiplicity_stats():
    stats1 = csr.summarize(csr.Skeleton(skeleton0), separator='_')
    stats2 = csr.summarize(csr.Skeleton(skeleton0, spacing=2), separator='_')
    assert_almost_equal(
            2 * stats1['branch_distance'].values,
            stats2['branch_distance'].values
            )
    assert_almost_equal(
            2 * stats1['euclidean_distance'].values,
            stats2['euclidean_distance'].values
            )


def test_pixel_values():
    image = np.random.random((45,))
    expected = np.mean(image)
    stats = csr.summarize(csr.Skeleton(image), separator='_')
    assert_almost_equal(stats.loc[0, 'mean_pixel_value'], expected)


def test_tip_junction_edges():
    stats1 = csr.summarize(csr.Skeleton(skeleton4), separator='_')
    assert stats1.shape[0] == 3  # ensure all three branches are counted


def test_mst_junctions():
    g, _ = csr.skeleton_to_csgraph(skeleton0)
    h = csr._mst_junctions(g)
    hprime, _ = csr.skeleton_to_csgraph(skeleton0)

    G = g.todense()
    G[G > 1.1] = 0

    np.testing.assert_equal(G, h.todense())
    np.testing.assert_equal(G, hprime.todense())


def test_transpose_image():
    image = np.zeros((10, 10))

    rr, cc = line(4, 0, 4, 2)
    image[rr, cc] = 1
    rr, cc = line(3, 2, 3, 5)
    image[rr, cc] = 1
    rr, cc = line(1, 2, 8, 2)
    image[rr, cc] = 1
    rr, cc = line(1, 0, 1, 8)
    image[rr, cc] = 1

    skeleton1 = csr.Skeleton(image)
    skeleton2 = csr.Skeleton(image.T)

    assert skeleton1.n_paths == skeleton2.n_paths
    np.testing.assert_allclose(
            np.sort(skeleton1.path_lengths()),
            np.sort(skeleton2.path_lengths()),
            )


@pytest.mark.parametrize(
        'skeleton,prune_branch,target',
        [
                (
                        skeleton1, 1,
                        np.array([[0, 1, 1, 1, 1, 1, 0], [1, 0, 0, 0, 0, 0, 1],
                                  [0, 1, 1, 0, 1, 1, 0], [0, 0, 0, 1, 0, 0, 0],
                                  [0, 0, 0, 0, 0, 0, 0]])
                        ),
                (
                        skeleton1, 2,
                        np.array([[0, 0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0, 0],
                                  [0, 1, 0, 0, 0, 0, 0], [1, 0, 0, 2, 0, 0, 0],
                                  [1, 0, 0, 0, 2, 2, 2]])
                        ),
                # There are no isolated cycles to be pruned
                (
                        skeleton1, 3,
                        np.array([[0, 1, 1, 1, 1, 1, 0], [1, 0, 0, 0, 0, 0, 1],
                                  [0, 3, 2, 0, 1, 1, 0], [3, 0, 0, 4, 0, 0, 0],
                                  [3, 0, 0, 0, 4, 4, 4]])
                        ),
                ],
        )
def test_prune_paths(
        skeleton: np.ndarray, prune_branch: int, target: np.ndarray
        ) -> None:
    """Test pruning of paths."""
    s = csr.Skeleton(skeleton, keep_images=True)
    summary = csr.summarize(s, separator='_')
    indices_to_remove = summary.loc[summary['branch_type'] == prune_branch
                                    ].index
    pruned = s.prune_paths(indices_to_remove)
    np.testing.assert_array_equal(pruned, target)


def test_prune_paths_exception_single_point() -> None:
    """Test exceptions raised when pruning leaves a single point and Skeleton object
    can not be created and returned."""
    s = csr.Skeleton(skeleton0)
    summary = csr.summarize(s, separator='_')
    indices_to_remove = summary.loc[summary['branch_type'] == 1].index
    with pytest.raises(ValueError):
        s.prune_paths(indices_to_remove)


def test_prune_paths_exception_invalid_path_index() -> None:
    """Test exceptions raised when trying to prune paths that do not exist in the summary. This can arise if skeletons
    are not updated correctly during iterative pruning."""
    s = csr.Skeleton(skeleton0)
    summary = csr.summarize(s, separator='_')
    indices_to_remove = [6]
    with pytest.raises(ValueError):
        s.prune_paths(indices_to_remove)


def test_fast_graph_center_idx():
    s = csr.Skeleton(skeleton0)
    i = csr._fast_graph_center_idx(s)
    assert i == 6

    s = csr.Skeleton(skeleton4)
    i = csr._fast_graph_center_idx(s)
    assert i == 1


def test_sholl():
    s = csr.Skeleton(skeleton0)
    c, r, counts = csr.sholl_analysis(s, shells=np.arange(0, 5, 1.5))
    np.testing.assert_equal(c, [3, 3])
    np.testing.assert_equal(counts, [0, 3, 3, 0])


def test_sholl_spacing():
    s = csr.Skeleton(skeleton0, spacing=(1, 5))
    with pytest.warns(UserWarning):
        c, r, counts = csr.sholl_analysis(
                s, center=[3, 15], shells=np.arange(17)
                )
        for i in range(4):
            assert np.isin(i, counts)
    c, r, counts = csr.sholl_analysis(
            s, center=[3, 15], shells=np.arange(1, 20, 6)
            )
    np.testing.assert_equal(counts, [3, 2, 2, 0])


def test_diagonal():
    s = csr.Skeleton(skeleton4)
    # We choose the shells so that we catch all three, then two, then one arm
    # of the skeleton, while not triggering the "shell spacing too small"
    # warning
    c, r, counts = csr.sholl_analysis(
            s, center=[1, 1], shells=np.arange(0.09, 5, 1.45)
            )
    np.testing.assert_equal(counts, [3, 2, 1, 0])


def test_zero_degree_nodes():
    """Test that graphs with 0-degree nodes don't have allocation errors.

    In skan commits 7498831 or prior, isolated pixels, which have degree 0,
    would count as *negative* values when counting the total number of edges,
    resulting in a buffer too small to hold the edge info.

    See issue jni/skan#182.
    """
    x = np.array([1, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 1]).astype(bool)

    # the try/except causes an early segfault if the buffer overflows
    try:
        s = csr.Skeleton(x)
    except Exception as e:
        print(e)

    assert s.n_paths == 2
    np.testing.assert_equal(
            np.ravel(s.coordinates), [0, 1, 3, 5, 7, 9, 11, 12]
            )


def test_skeleton_path_image_no_keep_image():
    """See #208: "Skeleton.path_label_image requires keep_images=True."

    Before PR #210, it was an implicit requirement of path_label_image
    to create a Skeleton with keep_images=True. However, we only needed
    the shape of the image. This makes sure that the method works even
    when keep_images=False.
    """
    s = csr.Skeleton(skeleton2, keep_images=False)
    pli = s.path_label_image()
    assert np.max(pli) == s.n_paths


def test_skeletonlabel():
    stats = csr.summarize(csr.Skeleton(skeletonlabel))
    assert stats['mean-pixel-value'].max() == skeletonlabel.max()
    assert stats['mean-pixel-value'].max() > 1


@pytest.mark.parametrize(
        'dtype', [
                ''.join([pre, 'int', suf])
                for pre, suf in product(['u', ''], ['8', '16', '32', '64'])
                ]
        )
def test_skeleton_integer_dtype(dtype):
    stats = csr.summarize(
            csr.Skeleton(skeletonlabel.astype(dtype)), separator='_'
            )
    assert stats['mean_pixel_value'].max() == skeletonlabel.max()
    assert stats['mean_pixel_value'].max() > 1


@pytest.mark.parametrize('dtype', [np.float32, np.float64])
def test_skeleton_all_float_dtypes(dtype):
    """Test that skeleton data types can be both float32 and float64."""
    horse = ~data.horse()
    skeleton_image = skeletonize(horse)
    dt = ndi.distance_transform_edt(horse)
    float_skel = (dt * skeleton_image).astype(dtype)
    _ = csr.Skeleton(float_skel)


def test_default_summarize_separator():
    with pytest.warns(np.exceptions.VisibleDeprecationWarning,
                      match='separator in column name'):
        stats = csr.summarize(csr.Skeleton(skeletonlabel))
    assert 'skeleton-id' in stats


def test_skeletonlabel():
    stats = csr.summarize(csr.Skeleton(skeletonlabel))
    assert stats['mean-pixel-value'].max() == skeletonlabel.max()
    assert stats['mean-pixel-value'].max() > 1


@pytest.mark.parametrize(('np_skeleton', 'spacing', 'dtype'),
                         product(
                                 [
                                         tinycycle, tinyline, skeleton0,
                                         skeleton1, skeleton2, skeleton3d
                                         ],
                                 [1.0, 2.0, (5.0, 2.5, 2.5)],
                                 [bool, np.float32, np.float64],
                                 ))
def test_pathgraph_skeleton_equiv(np_skeleton, spacing, dtype):
    if dtype is not bool:
        np_skeleton = (
                np_skeleton
                * np.random.random(np_skeleton.shape).astype(dtype)
                )
    if isinstance(spacing, tuple):
        spacing = spacing[:np_skeleton.ndim]  # truncate spacing to image ndim
    s = csr.Skeleton(np_skeleton, spacing=spacing)
    p = csr.PathGraph.from_image(np_skeleton, spacing=spacing)
    p2 = csr.PathGraph.from_graph(
            adj=p.adj,
            node_coordinates=p.node_coordinates,
            node_values=p.node_values,
            spacing=spacing,
            )
    ss = csr.summarize(s)
    sp = csr.summarize(p)
    sp2 = csr.summarize(p2)

    np.testing.assert_allclose(ss.to_numpy(), sp.to_numpy())
    np.testing.assert_allclose(sp.to_numpy(), sp2.to_numpy())


@pytest.mark.parametrize(
        ('np_skeleton', 'summary', 'nodes', 'edges'),
        [
                pytest.param(
                        tinycycle, None, 1, 1, id='tinycircle (no summary)'
                        ),
                pytest.param(tinyline, None, 2, 1, id='tinyline (no summary)'),
                pytest.param(
                        skeleton0,
                        csr.summarize(csr.Skeleton(skeleton0), separator='_'),
                        4,
                        3,
                        id='skeleton0 (with summary)'
                        ),
                pytest.param(
                        skeleton1, None, 4, 4, id='skeleton1 (no summary)'
                        ),
                pytest.param(
                        skeleton1,
                        csr.summarize(csr.Skeleton(skeleton1)),
                        4,
                        4,
                        id='skeleton1 (with summary)'
                        ),
                pytest.param(
                        skeleton2,
                        csr.summarize(csr.Skeleton(skeleton2)),
                        8,
                        8,
                        id='skeleton2 (with summary)'
                        ),
                pytest.param(
                        skeleton3d, None, 7, 7, id='skeleton3d (no summary)'
                        ),
                pytest.param(
                        skeleton_loop1,
                        None,
                        10,
                        10,
                        id='skeleton_loop1 (no summary)'
                        ),
                pytest.param(
                        skeleton_loop2,
                        None,
                        10,
                        10,
                        id='skeleton_loop2 (no summary)'
                        ),
                pytest.param(
                        skeleton_linear1,
                        None,
                        24,
                        24,
                        id='skeleton_linear1 (no summary)',
                        marks=pytest.mark.xfail(
                                sys.version_info[:2] == (3, 8),
                                reason='Incorrect edege discovery (#225)'
                                )
                        ),
                pytest.param(
                        skeleton_linear2,
                        None,
                        4,
                        3,
                        id='skeleton_linear2 (no summary)'
                        ),
                pytest.param(
                        skeleton_linear3,
                        None,
                        20,
                        17,
                        id='skeleton_linear3 (no summary)'
                        ),
                ],
        )
def test_skeleton_to_nx(
        np_skeleton: npt.NDArray, summary: pd.DataFrame | None, edges: int,
        nodes: int
        ) -> None:
    """Test creation of NetworkX Graph from skeletons arrays and summary."""
    skeleton = csr.Skeleton(np_skeleton)
    skan_nx = csr.skeleton_to_nx(skeleton, summary)
    assert skan_nx.number_of_nodes() == nodes
    assert skan_nx.number_of_edges() == edges


@pytest.mark.parametrize(
        ('np_skeleton', 'summary'),
        [
                pytest.param(
                        tinycycle,
                        csr.summarize(csr.Skeleton(tinycycle)),
                        id='tinycircle'
                        ),
                pytest.param(
                        tinyline,
                        csr.summarize(csr.Skeleton(tinyline)),
                        id='tinyline'
                        ),
                pytest.param(
                        skeleton0,
                        csr.summarize(csr.Skeleton(skeleton0)),
                        id='skeleton0'
                        ),
                pytest.param(
                        skeleton1,
                        csr.summarize(csr.Skeleton(skeleton1)),
                        id='skeleton1'
                        ),
                pytest.param(
                        skeleton3d,
                        csr.summarize(csr.Skeleton(skeleton3d)),
                        id='skeleton3d (no summary)'
                        ),
                pytest.param(
                        skeleton_loop1,
                        csr.summarize(csr.Skeleton(skeleton_loop1)),
                        id='skeleton_loop1'
                        ),
                pytest.param(
                        skeleton_loop2,
                        csr.summarize(csr.Skeleton(skeleton_loop2)),
                        id='skeleton_loop2'
                        ),
                pytest.param(
                        skeleton_linear1,
                        csr.summarize(csr.Skeleton(skeleton_linear1)),
                        id='skeleton_lienar1'
                        ),
                pytest.param(
                        skeleton_linear2,
                        csr.summarize(csr.Skeleton(skeleton_linear2)),
                        id='skeleton_linear2'
                        ),
                pytest.param(
                        skeleton_linear3,
                        csr.summarize(csr.Skeleton(skeleton_linear3)),
                        id='skeleton_linear3'
                        ),
                ],
        )
def test_nx_to_skeleton(
        np_skeleton: npt.NDArray,
        summary: pd.DataFrame | None,
        ) -> None:
    """Test creation of Skeleton from NetworkX Graph."""
    skeleton = csr.Skeleton(np_skeleton)
    skan_nx = csr.skeleton_to_nx(skeleton, summary)
    skeleton_nx = csr.nx_to_skeleton(skan_nx)
    np.testing.assert_array_equal(np_skeleton, skeleton_nx.skeleton_image)


@pytest.mark.parametrize(
        ('wrong_skeleton'),
        [
                pytest.param(skeleton0, id='Numpy Array.'),
                pytest.param(csr.Skeleton(skeleton0), id='Skeleton.'),
                pytest.param(nx_graph, id='NetworkX Graph without edges.'),
                pytest.param(
                        nx_graph_edges,
                        id='NetworkX Graph with points outside image.'
                        ),
                ],
        )
def test_nx_to_skeleton_attribute_error(wrong_skeleton: Any) -> None:
    """Test various errors are raised by nx_to_skeleton()."""
    with pytest.raises(Exception):
        csr.nx_to_skeleton(wrong_skeleton)


@pytest.mark.parametrize(
        ('skeleton'),
        [
                pytest.param(skeleton0, id='Numpy Array'),
                pytest.param(csr.Skeleton(skeleton0), id='Skeleton'),
                pytest.param(nx_graph, id='NetworkX Graph without edges.'),
                ],
        )
def test_csr_to_nbgraph_attribute_error(skeleton: Any) -> None:
    """Raise AttributeError if csr_to_nbgraph() passed incomplete objects."""
    with pytest.raises(AttributeError):
        csr.csr_to_nbgraph(skeleton)


@pytest.mark.parametrize(
        ('graph'),
        [
                pytest.param(
                        scipy.sparse.csr_matrix(skeleton0),
                        id='Sparse matrix directly from Numpy Array',
                        ),
                pytest.param(
                        scipy.sparse.csr_matrix(csr.Skeleton(skeleton0)),
                        id='Sparse matrix from csr.Skeleton',
                        ),
                ],
        )
def test_csr_to_nbgraph_type_error(graph: scipy.sparse.csr_matrix) -> None:
    """Test TypeError is raised by csr_to_nbgraph() if wrong type is passed."""
    with pytest.raises(TypeError):
        csr.csr_to_nbgraph(graph)



import os

from contextlib import contextmanager
from collections import OrderedDict
from time import process_time

import numpy as np

from skan import csr

rundir = os.path.dirname(__file__)


@contextmanager
def timer():
    time = []
    t0 = process_time()
    yield time
    t1 = process_time()
    time.append(t1 - t0)


def bench_suite():
    times = OrderedDict()
    skeleton = np.load(os.path.join(rundir, 'infected3.npz'))['skeleton']
    with timer() as t_build_graph:
        g, indices = csr.skeleton_to_csgraph(skeleton, spacing=2.24826)
    times['build graph'] = t_build_graph[0]
    with timer() as t_build_graph2:
        g, indices = csr.skeleton_to_csgraph(skeleton, spacing=2.24826)
    times['build graph again'] = t_build_graph2[0]
    with timer() as t_stats:
        stats = csr.branch_statistics(g)
    times['compute statistics'] = t_stats[0]
    with timer() as t_stats2:
        stats = csr.branch_statistics(g)
    times['compute statistics again'] = t_stats2[0]
    with timer() as t_skeleton:
        skel_obj = csr.Skeleton(skeleton)
    times['skeleton object'] = t_skeleton[0]
    with timer() as t_skeleton2:
        skel_obj = csr.Skeleton(skeleton)
    times['skeleton object again'] = t_skeleton2[0]
    with timer() as t_summary:
        summary = csr.summarize(skel_obj, separator='_')
    times['compute per-skeleton statistics'] = t_summary[0]
    return times


def print_bench_results(times=None, memory=None):
    if times is not None:
        print('Timing results:')
        for key in times:
            print('--- ', key, '%.3f s' % times[key])
    if memory is not None:
        print('Memory results:')
        for key in memory:
            print('--- ', key, '%.3f MB' % (memory[key] / 1e6))


if __name__ == '__main__':
    times = bench_suite()
    print_bench_results(times)



import numpy as np
from scipy import spatial, ndimage as ndi
from skimage import filters, restoration
from skimage.filters import threshold_sauvola, threshold_niblack

SMOOTH_METHODS = {
        'Gaussian': filters.gaussian,
        'TV': restoration.denoise_tv_bregman,
        }


def hyperball(ndim, radius):
    """Return a binary morphological filter containing pixels within `radius`.

    Parameters
    ----------
    ndim : int
        The number of dimensions of the filter.
    radius : int
        The radius of the filter.

    Returns
    -------
    ball : array of bool, shape [2 * radius + 1,] * ndim
        The required structural element
    """
    size = 2*radius + 1
    center = [(radius,) * ndim]

    coords = np.mgrid[[
            slice(None, size),
            ] * ndim].reshape(ndim, -1).T
    distances = np.ravel(spatial.distance_matrix(coords, center))
    selector = distances <= radius

    ball = np.zeros((size,) * ndim, dtype=bool)
    ball.ravel()[selector] = True
    return ball


def threshold(
        image,
        *,
        sigma=0.,
        radius=0,
        offset=0.,
        method='sauvola',
        smooth_method='Gaussian'
        ):
    """Use scikit-image filters to "intelligently" threshold an image.

    Parameters
    ----------
    image : array, shape (M, N, ...[, 3])
        Input image, conformant with scikit-image data type
        specification [1]_.
    sigma : float, optional
        If positive, use Gaussian filtering to smooth the image before
        thresholding.
    radius : int, optional
        If given, use local median thresholding instead of global.
    offset : float, optional
        If given, reduce the threshold by this amount. Higher values
        result in fewer pixels above the threshold.
    method: {'sauvola', 'niblack', 'median'}
        Which method to use for thresholding. Sauvola is 100x faster, but
        median might be more accurate.
    smooth_method: {'Gaussian', 'TV'}
        Which method to use for smoothing. Choose from Gaussian smoothing
        and total variation denoising.

    Returns
    -------
    thresholded : image of bool, same shape as `image`
        The thresholded image.

    References
    ----------
    .. [1] http://scikit-image.org/docs/dev/user_guide/data_types.html
    """
    if sigma > 0:
        if smooth_method.lower() == 'gaussian':
            image = filters.gaussian(image, sigma=sigma)
        elif smooth_method.lower() == 'tv':
            image = restoration.denoise_tv_bregman(image, weight=sigma)
    if radius == 0:
        t = filters.threshold_otsu(image) + offset
    else:
        if method == 'median':
            footprint = hyperball(image.ndim, radius=radius)
            t = ndi.median_filter(image, footprint=footprint) + offset
        elif method == 'sauvola':
            w = 2*radius + 1
            t = threshold_sauvola(image, window_size=w, k=offset)
        elif method == 'niblack':
            w = 2*radius + 1
            t = threshold_niblack(image, window_size=w, k=offset)
        else:
            raise ValueError(
                    'Unknown method %s. Valid methods are median,'
                    'niblack, and sauvola.' % method
                    )
    thresholded = image > t
    return thresholded



from skan.napari_skan import labels_to_skeleton_shapes, _update_feature_names
from skimage import data, morphology
from napari.layers import Labels
import numpy as np
from skan.napari_skan import SkeletonizeMethod
from skan.csr import Skeleton
import pandas as pd
import napari


def make_trivial_labels_layer():
    label_data = np.zeros(shape=(10, 10), dtype=int)
    label_data[5, 1:9] = 1
    labels_layer = Labels(label_data)
    return labels_layer


def test_get_skeleton_simple():
    labels_layer = make_trivial_labels_layer()
    skeleton_type = SkeletonizeMethod.zhang
    shapes_data, layer_kwargs, _ = labels_to_skeleton_shapes(
            labels_layer, skeleton_type
            )

    assert type(layer_kwargs['metadata']['skeleton']) is Skeleton
    np.testing.assert_array_equal(
            shapes_data[0],
            [[5, 1], [5, 2], [5, 3], [5, 4], [5, 5], [5, 6], [5, 7], [5, 8]]
            )
    assert len(shapes_data) == 1
    assert 'features' in layer_kwargs
    assert type(layer_kwargs['features']) is pd.DataFrame


def test_get_skeleton_horse():
    horse = np.logical_not(data.horse().astype(bool))
    labels_layer = Labels(horse)
    skeleton_type = SkeletonizeMethod.zhang
    shapes_data, layer_kwargs, _ = labels_to_skeleton_shapes(
            labels_layer, skeleton_type
            )
    assert len(shapes_data) == 24  # 24 line segments in the horse skeleton
    assert 'features' in layer_kwargs
    assert type(layer_kwargs['features']) is pd.DataFrame


def test_gui(make_napari_viewer):
    viewer = make_napari_viewer()
    horse = np.logical_not(data.horse().astype(bool))

    labels_layer = viewer.add_labels(horse)

    ldt = labels_to_skeleton_shapes(labels_layer, SkeletonizeMethod.zhang)
    (skel_layer,) = viewer._add_layer_from_data(*ldt)

    dw, widget = viewer.window.add_plugin_dock_widget(
            'skan', 'Color Skeleton Widget'
            )
    widget.feature_name.value = 'euclidean_distance'
    widget()
    layer = viewer.layers[-1]
    assert layer.edge_colormap.name == 'viridis'
    assert len(layer.data) == 24



import os
import imageio
from scipy import special
import numpy as np
from hypothesis import given, strategies, settings
import pytest

from skan import pre


@pytest.fixture
def image():
    rundir = os.path.abspath(os.path.dirname(__file__))
    datadir = os.path.join(rundir, 'data')
    image = imageio.v2.imread(os.path.join(datadir, 'retic.tif'), format='fei')
    return image


def ball_volume(ndim, radius):
    """Return the volume of a ball of dimension `ndim` and radius `radius`."""
    n = ndim
    r = radius
    return np.pi**(n / 2) / special.gamma(n/2 + 1) * r**n


@settings(deadline=None)
@given(
        strategies.integers(min_value=1, max_value=4),
        strategies.integers(min_value=2, max_value=10),
        )
def test_hyperball_volume(ndim, radius):
    theoretical_volume = ball_volume(ndim, radius)
    approximate_volume = np.sum(pre.hyperball(ndim, radius))
    np.testing.assert_allclose(
            approximate_volume, theoretical_volume, rtol=0.5
            )


def test_threshold2d_sauvola(image):
    res = image.meta['Scan']['PixelHeight'] * 1e9  # nm/pixel
    radius = int(50 / res)  # radius of 50nm in pixels
    sigma = 0.1 * radius
    thresholded0 = pre.threshold(
            image, sigma=sigma, radius=radius, method='sauvola', offset=0.2
            )
    assert thresholded0.dtype == bool
    assert thresholded0.shape == image.shape
    thresholded1 = pre.threshold(
            image, sigma=sigma, radius=radius, method='sauvola', offset=0.075
            )
    assert np.all(thresholded1 <= thresholded0)


def test_threshold_2d_niblack(image):
    res = image.meta['Scan']['PixelHeight'] * 1e9  # nm/pixel
    radius = int(50 / res)
    sigma = 0.1 * radius
    thresholded0 = pre.threshold(
            image, sigma=sigma, radius=radius, method='niblack', offset=0.075
            )
    assert thresholded0.shape == image.shape


def test_threshold_2d_median(image):
    sigma = 2
    radius = 5
    thresholded = pre.threshold(
            image[:100, :100], sigma=sigma, radius=radius, method='median'
            )
    assert thresholded.shape == (100, 100)
    assert thresholded.dtype == bool


def test_threshold_2d_otsu(image):
    thresholded_otsu = pre.threshold(image)
    assert thresholded_otsu.shape == image.shape


def test_threshold_no_method(image):
    with pytest.raises(ValueError):
        pre.threshold(image, radius=1, method='no method')


def _total_variation(image):
    return sum(
            np.sum(np.abs(np.diff(image, axis=i))) for i in range(image.ndim)
            )


def test_threshold_denoise(image):
    denoised_thresholded = pre.threshold(
            image, sigma=5, radius=15, smooth_method='tv'
            )
    thresholded = pre.threshold(image, sigma=0, radius=15)
    assert (
            _total_variation(thresholded) >
            _total_variation(denoised_thresholded)
            )



import networkx as nx
import numpy as np
from skimage.draw import random_shapes
from skimage.morphology import skeletonize

tinycycle = np.array([[0, 1, 0],
                      [1, 0, 1],
                      [0, 1, 0]], dtype=bool)


tinyline = np.array([0, 1, 1, 1, 0], dtype=bool)


skeleton0 = np.array([[0, 0, 0, 1, 0, 0, 0],
                      [0, 0, 0, 1, 0, 0, 0],
                      [0, 0, 0, 1, 0, 0, 0],
                      [1, 1, 1, 1, 1, 1, 1]], dtype=bool)


skeleton1 = np.array([[0, 1, 1, 1, 1, 1, 0],
                      [1, 0, 0, 0, 0, 0, 1],
                      [0, 1, 1, 0, 1, 1, 0],
                      [1, 0, 0, 1, 0, 0, 0],
                      [1, 0, 0, 0, 1, 1, 1]], dtype=bool)


_zeros1 = np.zeros_like(skeleton1)
skeleton2 = np.concatenate((skeleton1, _zeros1), axis=1)
skeleton2 = np.concatenate((skeleton2, skeleton2[:, ::-1]), axis=0)

skeleton3d = np.array([[[1, 0, 0, 0, 0],
                        [0, 0, 0, 0, 0],
                        [0, 0, 0, 0, 0],
                        [0, 0, 0, 0, 0],
                        [0, 0, 0, 0, 0]],
                       [[0, 0, 0, 0, 0],
                        [0, 1, 0, 0, 1],
                        [0, 0, 0, 0, 1],
                        [0, 0, 1, 0, 1],
                        [1, 1, 0, 1, 0]],
                       [[0, 0, 0, 1, 0],
                        [0, 0, 0, 0, 0],
                        [0, 0, 1, 0, 0],
                        [0, 0, 0, 0, 0],
                        [0, 0, 0, 0, 0]],
                       [[0, 0, 0, 0, 0],
                        [0, 0, 0, 1, 0],
                        [0, 0, 0, 0, 0],
                        [0, 0, 0, 0, 0],
                        [0, 1, 0, 0, 0]],
                       [[0, 0, 0, 0, 0],
                        [0, 0, 0, 0, 0],
                        [0, 1, 0, 0, 1],
                        [1, 0, 1, 0, 1],
                        [0, 0, 0, 0, 1]]], dtype=bool)

topograph1d = np.array([3., 2., 3.])

skeleton4 = np.array([[1, 0, 0, 0, 0],
                      [0, 1, 1, 1, 1],
                      [0, 1, 0, 0, 0],
                      [0, 1, 0, 0, 0]], dtype=bool)

junction_first = np.array([[0, 1, 1, 1, 1],
                           [1, 1, 0, 0, 0],
                           [1, 0, 1, 0, 0],
                           [1, 0, 0, 1, 0],
                           [1, 0, 0, 0, 1]], dtype=bool)

skeletonlabel = np.array([[1, 1, 0, 0, 2, 2, 0],
                          [0, 0, 1, 0, 0, 0, 2],
                          [3, 0, 0, 1, 0, 0, 2],
                          [3, 0, 0, 1, 0, 0, 0],
                          [3, 0, 0, 0, 1, 1, 1],
                          [0, 3, 0, 0, 0, 1, 0]], dtype=int)


def _generate_random_skeleton(**extra_kwargs):
    """Generate random skeletons using skimage.draw's random_shapes."""
    kwargs = {"image_shape": (128, 128),
              "max_shapes": 20,
              "channel_axis": None,
              "shape": None,
              "allow_overlap": True}
    random_image, _ = random_shapes(**kwargs, **extra_kwargs)
    mask = random_image != 255
    return skeletonize(mask)

# Generate random skeletons:

# Skeleton with loop to be retained and side-branches
skeleton_loop1 = _generate_random_skeleton(rng=1, min_size=20)
# Skeleton with loop to be retained and side-branches
skeleton_loop2 = _generate_random_skeleton(rng=165103, min_size=60)
# Linear skeleton with lots of large side-branches, some forked
skeleton_linear1 = _generate_random_skeleton(rng=13588686514, min_size=20)
# Linear Skeleton with simple fork at one end
skeleton_linear2 = _generate_random_skeleton(rng=21, min_size=20)
# Linear Skeletons (i.e. multiple) with branches
skeleton_linear3 = _generate_random_skeleton(rng=894632511, min_size=20)

## Sample NetworkX Graphs...
# ...with no edge attributes
nx_graph = nx.Graph()
nx_graph.add_nodes_from([1, 2, 3])
# ...with edge attributes
nx_graph_edges = nx.Graph()
nx_graph_edges.add_nodes_from([1, 2, 3])
nx_graph_edges.add_edge(1, 2, **{"path": np.asarray([[4, 4]])})



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Skan documentation build configuration file, created by
# sphinx-quickstart on Tue Dec 12 15:07:05 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# Add the local code to the Python path, so docs are generated for
# current working copy
import os
import re
import sys

rundir = os.path.dirname(__file__)
sys.path.insert(0, os.path.abspath(rundir + '/..'))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Execution options
execution_timeout = 300  # in seconds?

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
        'sphinx.ext.autodoc',
        'sphinx.ext.mathjax',
        'sphinx.ext.viewcode',
        'sphinx.ext.githubpages',
        'sphinx.ext.napoleon',
        'sphinxcontrib.bibtex',
        'sphinx_copybutton',
        'myst_nb',
        ]

# sphinx-copybutton prompt-stripping configuration
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Skan 🦴📏'
copyright = '2021, Juan Nunez-Iglesias & skan contributors'
author = 'Juan Nunez-Iglesias'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
import skan

version = skan.__version__
# The full version, including alpha/beta/rc tags.
release = skan.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# List of publications
bibtex_bibfiles = ['publications.bib']

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

html_theme = 'pydata_sphinx_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#

# Version match must match the 'version' key in version_swticher.json
pattern = re.compile(r'^[0-9]+\.[0-9]+')
version_match = pattern.search(version)
if 'dev' in version:
    version_match = version
elif version_match:
    version_match = version_match.group() + ".x"

# PyData theme options (see theme.conf for more information)
html_theme_options = {
    # Set the name of the project to appear in the navigation.
    'navbar_end': ['version-switcher', 'navbar-icon-links'],
    # pydata theme version switcher config
    'switcher': {
        'json_url': 'https://skeleton-analysis.org/dev/_static/version_switcher.json',
        'version_match': version_match,
    },
    # footer
    'footer_items': ['copyright'],
    'icon_links': [{'name': 'GitHub', 'url': 'https://github.com/jni/skan',
            'icon': 'fa-brands fa-square-github', 'type': 'fontawesome'}]
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'Skandoc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
        # The paper size ('letterpaper' or 'a4paper').
        #
        # 'papersize': 'letterpaper',

        # The font size ('10pt', '11pt' or '12pt').
        #
        # 'pointsize': '10pt',

        # Additional stuff for the LaTeX preamble.
        #
        # 'preamble': '',

        # Latex figure (float) alignment
        #
        # 'figure_align': 'htbp',
        }

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
        (
                master_doc, 'Skan.tex', 'Skan Documentation',
                'Juan Nunez-Iglesias', 'manual'
                ),
        ]

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, 'skan', 'Skan Documentation', [author], 1)]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
        (
                master_doc, 'Skan', 'Skan Documentation', author, 'Skan',
                'Skeleton analysis in Python.', 'Miscellaneous'
                ),
        ]



import numpy as np

from skan._testdata import skeleton1
from skan import Skeleton, summarize


def test_find_main():
    skeleton = Skeleton(skeleton1)
    summary_df = summarize(skeleton, find_main_branch=True, separator='_')

    non_main_edge_start = [2, 1]
    non_main_edge_finish = [3, 3]

    non_main_df = summary_df.loc[summary_df['main'] == False]
    assert non_main_df.shape[0] == 1
    coords = non_main_df[[
            'coord_src_0', 'coord_src_1', 'coord_dst_0', 'coord_dst_1'
            ]].to_numpy()
    assert (
            np.all(coords == non_main_edge_start + non_main_edge_finish)
            or np.all(coords == non_main_edge_finish + non_main_edge_start)
            )



from magicgui import magic_factory
import numpy as np
from enum import Enum
from skimage.morphology import skeletonize
from skan import summarize, Skeleton

CAT_COLOR = "tab10"
CONTINUOUS_COLOR = "viridis"


class SkeletonizeMethod(Enum):
    """Use enum for method choice for easier use with magicgui."""
    zhang = "zhang"
    lee = "lee"


def labels_to_skeleton_shapes(
        labels: "napari.layers.Labels", choice: SkeletonizeMethod
        ) -> "napari.types.LayerDataTuple":
    """Skeletonize a labels layer using given method and export as Shapes.

    Parameters
    ----------
    labels : napari.layers.Labels
        Labels layer containing data to skeletonize
    choice : SkeletonizeMethod
        Enum corresponding to skeletonization method

    Returns
    -------
    napari.types.LayerDataTuple
        Shapes layer data with skeleton
    """
    binary_labels = (labels.data > 0).astype(np.uint8)
    binary_skeleton = skeletonize(binary_labels, method=choice.value)

    skeleton = Skeleton(binary_skeleton)

    all_paths = [skeleton.path_coordinates(i) for i in range(skeleton.n_paths)]

    # option to have main_path = True (or something) changing header
    paths_table = summarize(skeleton, separator='_')
    layer_kwargs = {
            'shape_type': 'path',
            'edge_colormap': 'tab10',
            'features': paths_table,
            'metadata': {'skeleton': skeleton},
            }

    return all_paths, layer_kwargs, 'shapes'


def _populate_feature_choices(color_by_feature_widget):
    """Update feature names combobox when source layer is changed.

    This runs on widget initialization and on every change of Shapes layer
    thereafter.

    Parameters
    ----------
    color_by_feature_widget : function that takes widget as input
        Function that takes in the widget and modifies the choices in-place.
    """
    color_by_feature_widget.shapes_layer.changed.connect(
            lambda _: _update_feature_names(color_by_feature_widget)
            )
    _update_feature_names(color_by_feature_widget)


def _update_feature_names(color_by_feature_widget):
    """Search for a shapes layer with appropriate metadata for skeletons

    Parameters
    ----------
    color_by_feature_widget : magicgui Widget
        widget that contains reference to shapes layers
    """
    shapes_layer = color_by_feature_widget.shapes_layer.value

    def get_choices(features_combo):
        """Closure to use the current shapes layer to update given combobox."""
        return shapes_layer.features.columns

    color_by_feature_widget.feature_name.choices = get_choices


@magic_factory(
        widget_init=_populate_feature_choices,
        feature_name={"widget_type": "ComboBox"}
        )
def color_by_feature(shapes_layer: "napari.layers.Shapes", feature_name):
    """Check the currently selected feature and update edge colors

    TODO: allow selecting a colormap.

    Parameters
    ----------
    shapes_layer : napari.layers.Shapes
        A napari Shapes layer.
    feature_name : String
        A string corresponding to a feature column present in the Shapes layer.
    """
    current_column_type = shapes_layer.features[feature_name].dtype
    if current_column_type == "float64":
        shapes_layer.edge_colormap = CONTINUOUS_COLOR
    else:
        shapes_layer.edge_colormap = CAT_COLOR
    shapes_layer.edge_color = feature_name



import os
from . import pre, csr
import imageio
from tqdm import tqdm
import numpy as np
from skimage import morphology
import pandas as pd
from .image_stats import image_summary
from skimage.feature import shape_index
from concurrent.futures import ThreadPoolExecutor, as_completed
import multiprocessing as mp

CPU_COUNT = int(os.environ.get('CPU_COUNT', mp.cpu_count()))


def _get_scale(image, md_path_or_scale):
    """Get a valid scale from an image and a metadata path or scale.

    Parameters
    ----------
    image : np.ndarray
        The input image.
    md_path_or_scale : float or image filename
        The path to the file containing the metadata, or the scale.

    Returns
    -------
    scale : float
    """
    scale = None
    try:
        scale = float(md_path_or_scale)
    except ValueError:
        pass
    if md_path_or_scale is not None and scale is None:
        md_path = md_path_or_scale.split(sep='/')
        meta = image.meta
        for key in md_path:
            meta = meta[key]
        scale = float(meta)
    else:
        if scale is None:
            scale = 1  # measurements will be in pixel units
    return scale


def process_single_image(
        filename, image_format, scale_metadata_path, threshold_radius,
        smooth_radius, brightness_offset, crop_radius, smooth_method
        ):
    image = imageio.v2.imread(filename, format=image_format)
    scale = _get_scale(image, scale_metadata_path)
    if crop_radius > 0:
        c = crop_radius
        image = image[c:-c, c:-c]
    pixel_threshold_radius = int(np.ceil(threshold_radius / scale))

    pixel_smoothing_radius = smooth_radius * pixel_threshold_radius
    thresholded = pre.threshold(
            image,
            sigma=pixel_smoothing_radius,
            radius=pixel_threshold_radius,
            offset=brightness_offset,
            smooth_method=smooth_method
            )
    quality = shape_index(image, sigma=pixel_smoothing_radius, mode='reflect')
    skeleton = morphology.skeletonize(thresholded) * quality
    framedata = csr.summarize(
            csr.Skeleton(skeleton, spacing=scale), separator='_'
            )
    framedata['squiggle'] = np.log2(
            framedata['branch_distance'] / framedata['euclidean_distance']
            )
    framedata['scale'] = scale
    framedata.rename(
            columns={'mean_pixel_value': 'mean_shape_index'},
            inplace=True,
            errors='raise',
            )
    framedata['filename'] = filename
    return image, thresholded, skeleton, framedata


def process_images(
        filenames,
        image_format,
        threshold_radius,
        smooth_radius,
        brightness_offset,
        scale_metadata_path,
        crop_radius=0,
        smooth_method='Gaussian',
        num_threads=CPU_COUNT
        ):
    """Full pipeline from images to skeleton stats with local median threshold.

    Parameters
    ----------
    filenames : list of string
        The list of input filenames.
    image_format : string
        The format of the files. 'auto' is automatically determined by the
        imageio library. See imageio documentation for valid image formats.
    threshold_radius : float
        The radius for median thresholding,
    smooth_radius : float in [0, 1]
        The value of sigma with which to Gaussian-smooth the image,
        **relative to `threshold_radius`**.
    brightness_offset : float
        The standard brightness value with which to threshold is the local
        median, `m(x, y)`. Use this value to offset from there: the threshold
        used will be `m(x, y) + brightness_offset`.
    scale_metadata_path : string
        The path in the image dictionary to find the metadata on pixel scale,
        separated by forward slashes ('/').
    crop_radius : int, optional
        Crop `crop_radius` pixels from each margin of the image before
        processing.
    smooth_method : {'Gaussian', 'TV', 'NL'}, optional
        Which method to use for smoothing.
    num_threads : int, optional
        How many threads to use for computation. This should generally be
        set to the number of CPU cores available to you.

    Returns
    -------
    results : generator
        The pipeline yields individual image results in the form of a tuple
        of ``(filename, image, thresholded_image, skeleton, data_frame)``.
        Finally, after all the images have been processed, the pipeline yields
        a DataFrame containing all the collated branch-level results.
    """
    image_format = None if image_format == 'auto' else image_format
    results = []
    image_results = []
    with ThreadPoolExecutor(max_workers=num_threads) as ex:
        future_data = {
                ex.submit(
                        process_single_image, filename, image_format,
                        scale_metadata_path, threshold_radius, smooth_radius,
                        brightness_offset, crop_radius, smooth_method
                        ): filename
                for filename in filenames
                }
        for completed_data in tqdm(as_completed(future_data),
                                   total=len(filenames)):
            image, thresholded, skeleton, framedata = completed_data.result()
            filename = future_data[completed_data]
            results.append(framedata)
            image_stats = image_summary(
                    skeleton, spacing=framedata['scale'][0]
                    )
            image_stats['filename'] = filename
            image_stats['branch density'] = (
                    framedata.shape[0] / image_stats['area']
                    )
            j2j = framedata[framedata['branch_type'] == 2]
            image_stats['mean J2J branch distance'] = (
                    j2j['branch_distance'].mean()
                    )
            image_results.append(image_stats)
            yield filename, image, thresholded, skeleton, framedata
    yield pd.concat(results), pd.concat(image_results)



import pandas as pd


def _params_dict_to_dataframe(d):
    s = pd.Series(d)
    s.index.name = 'parameters'
    f = pd.DataFrame({'values': s})
    return f


def write_excel(filename, **kwargs):
    """Write data tables to an Excel file, using kwarg names as sheet names.
    
    Parameters
    ----------
    filename : str
        The filename to write to.
    kwargs : dict
        Mapping from sheet names to data.
    """
    writer = pd.ExcelWriter(filename)
    for sheet_name, obj in kwargs.items():
        if isinstance(obj, dict):
            obj = _params_dict_to_dataframe(obj)
        if isinstance(obj, pd.DataFrame):
            obj.to_excel(writer, sheet_name=sheet_name)
    writer.close()
